from snowddl.blueprint import (
    AccountGrant,
    AccountObjectIdent,
    DatabaseBlueprint,
    DatabaseIdent,
    Ident,
    IdentPattern,
    build_share_read_ident,
)
from snowddl.parser.abc_parser import AbstractParser


# fmt: off
database_json_schema = {
    "type": "object",
    "properties": {
        "permission_model": {
            "type": "string",
        },
        "is_transient": {
            "type": "boolean"
        },
        "retention_time": {
            "type": "integer"
        },
        "external_volume": {
            "type": "string"
        },
        "catalog": {
            "type": "string",
        },
        "is_sandbox": {
            "type": "boolean"
        },
        "owner_database_read": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "owner_database_write": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "owner_schema_read": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "owner_schema_write": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "owner_share_read": {
            "share_read": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
        },
        "owner_integration_usage": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "owner_warehouse_usage": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "owner_account_grants": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "owner_global_roles": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "comment": {
            "type": "string"
        }
    },
    "additionalProperties": False
}
# fmt: on


class DatabaseParser(AbstractParser):
    def load_blueprints(self):
        for database_name in self.get_database_names():
            database_params = self.parse_single_entity_file(f"{database_name}/params", database_json_schema)
            database_permission_model_name = database_params.get("permission_model", self.config.DEFAULT_PERMISSION_MODEL).upper()

            # fmt: off
            bp = DatabaseBlueprint(
                full_name=DatabaseIdent(self.env_prefix, database_name),
                permission_model=database_permission_model_name,
                is_transient=database_params.get("is_transient", False),
                retention_time=database_params.get("retention_time", None),
                external_volume=Ident(database_params.get("external_volume")) if database_params.get("external_volume") else None,
                catalog=Ident(database_params.get("catalog")) if database_params.get("catalog") else None,
                is_sandbox=database_params.get("is_sandbox", False),
                owner_database_write=[IdentPattern(p) for p in database_params.get("owner_database_write", [])],
                owner_database_read=[IdentPattern(p) for p in database_params.get("owner_database_read", [])],
                owner_schema_write=[IdentPattern(p) for p in database_params.get("owner_schema_write", [])],
                owner_schema_read=[IdentPattern(p) for p in database_params.get("owner_schema_read", [])],
                owner_integration_usage=[Ident(integration_name) for integration_name in database_params.get("owner_integration_usage", [])],
                owner_share_read=[build_share_read_ident(share_name) for share_name in database_params.get("owner_share_read", [])],
                owner_warehouse_usage=[AccountObjectIdent(self.env_prefix, warehouse_name) for warehouse_name in database_params.get("owner_warehouse_usage", [])],
                owner_account_grants=[AccountGrant(privilege=privilege) for privilege in database_params.get("owner_account_grants", [])],
                owner_global_roles=[Ident(global_role_name) for global_role_name in database_params.get("owner_global_roles", [])],
                comment=database_params.get("comment", None),
            )
            # fmt: on

            self.config.add_blueprint(bp)



from snowddl.blueprint import BackupSetBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class BackupSetResolver(AbstractSchemaObjectResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.BACKUP_SET

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW BACKUP SETS IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            object_type = r["object_kind"]  # more fancy logic is going to be added here

            if object_type == "DATABASE":
                object_name = r["object_name"]
            elif object_type == "SCHEMA":
                object_name = f"{r['object_database_name']}.{r['object_name']}"
            else:
                object_name = f"{r['object_database_name']}.{r['object_schema_name']}.{r['object_name']}"

            if r["backup_policy_name"]:
                backup_policy = f"{r['backup_policy_database_name']}.{r['backup_policy_schema_name']}.{r['backup_policy_name']}"
            else:
                backup_policy = None

            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner_role"],
                "object_type": object_type,
                "object_name": object_name,
                "backup_policy": backup_policy,
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(BackupSetBlueprint)

    def create_object(self, bp: BackupSetBlueprint):
        query = self.engine.query_builder()

        query.append(
            "CREATE BACKUP SET {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        query.append_nl(
            "FOR {object_type:r} {object_name:i}",
            {
                "object_type": bp.object_type.singular,
                "object_name": bp.object_name,
            },
        )

        if bp.backup_policy:
            query.append_nl(
                "WITH BACKUP POLICY {backup_policy:i}",
                {
                    "backup_policy": bp.backup_policy,
                },
            )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        self.engine.execute_unsafe_ddl(query)

        return ResolveResult.CREATE

    def compare_object(self, bp: BackupSetBlueprint, row: dict):
        result = ResolveResult.NOCHANGE
        replace_reasons = []

        if bp.backup_policy is None and row["backup_policy"] is not None:
            replace_reasons.append("Cannot remove backup policy from backup set")

        if bp.object_type.singular != row["object_type"]:
            replace_reasons.append("Object type for backup set was changed")

        if bp.object_name != row["object_name"]:
            replace_reasons.append("Object name for backup set was changed")

        if replace_reasons:
            query = self.engine.query_builder()

            query.append(
                "CREATE OR REPLACE BACKUP SET {full_name:i}",
                {
                    "full_name": bp.full_name,
                },
            )

            query.append_nl(
                "FOR {object_type:r} {object_name:i}",
                {
                    "object_type": bp.object_type.singular,
                    "object_name": bp.object_name,
                },
            )

            if bp.backup_policy:
                query.append_nl(
                    "WITH BACKUP POLICY {backup_policy:i}",
                    {
                        "backup_policy": bp.backup_policy,
                    },
                )

            if bp.comment:
                query.append_nl(
                    "COMMENT = {comment}",
                    {
                        "comment": bp.comment,
                    },
                )

            self.engine.execute_unsafe_ddl(query)

            return ResolveResult.REPLACE

        if bp.backup_policy != row["backup_policy"]:
            if bp.backup_policy:
                self.engine.execute_unsafe_ddl(
                    "ALTER BACKUP SET {full_name:i} APPLY BACKUP POLICY {backup_policy:i} FORCE",
                    {
                        "full_name": bp.full_name,
                        "backup_policy": bp.backup_policy,
                    },
                )

                result = ResolveResult.ALTER

        if bp.comment != row["comment"]:
            self.engine.execute_unsafe_ddl(
                "ALTER BACKUP SET {full_name:i} SET COMMENT = {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": bp.comment,
                },
            )

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP BACKUP SET {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        return ResolveResult.DROP



from json import loads as json_loads
from typing import List

from snowddl.blueprint import AccountIdent, OutboundShareBlueprint, Grant, GrantPattern, build_grant_name_ident
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType


class OutboundShareResolver(AbstractResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.SHARE

    def get_existing_objects(self):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW SHARES")

        for r in cur:
            if r["kind"] != "OUTBOUND":
                continue

            if r["owner"] != self.engine.context.current_role:
                continue

            # Remove organization and account prefix, shares are referred in SQL by name only
            full_name = r["name"].split(".")[-1]

            existing_objects[full_name] = {
                "share": full_name,
                "database": r["database_name"],
                "accounts": r["to"].split(",") if r["to"] else [],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(OutboundShareBlueprint)

    def create_object(self, bp: OutboundShareBlueprint):
        query = self.engine.query_builder()

        query.append(
            "CREATE SHARE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_outbound_share)

        for bp_grant in self.expand_grant_patterns_to_grants(bp.grant_patterns):
            self.create_grant(bp.full_name, bp_grant)

        self.compare_accounts(bp)

        return ResolveResult.CREATE

    def compare_object(self, bp: OutboundShareBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        bp_grants = self.expand_grant_patterns_to_grants(bp.grant_patterns)
        existing_grants = self.get_existing_share_grants(bp.full_name)

        for bp_grant in bp_grants:
            if bp_grant not in existing_grants:
                self.create_grant(bp.full_name, bp_grant)
                result = ResolveResult.GRANT

        for ex_grant in existing_grants:
            if not any(grant_pattern.is_matching_grant(ex_grant) for grant_pattern in bp.grant_patterns):
                self.drop_grant(bp.full_name, ex_grant)
                result = ResolveResult.GRANT

        if self.compare_accounts(bp, check_existing=True):
            result = ResolveResult.ALTER

        if bp.comment != row["comment"]:
            self.engine.execute_safe_ddl(
                "ALTER SHARE {full_name:i} SET COMMENT = {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": bp.comment,
                },
                condition=self.engine.settings.execute_outbound_share,
            )

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP SHARE {share_name:i}",
            {
                "share_name": row["share"],
            },
            condition=self.engine.settings.execute_outbound_share,
        )

        return ResolveResult.DROP

    def compare_accounts(self, bp: OutboundShareBlueprint, check_existing=False):
        existing_accounts = []
        accounts_to_add = []
        accounts_to_remove = []

        if check_existing:
            cur = self.engine.execute_meta(
                "SELECT SYSTEM$LIST_OUTBOUND_SHARES_DETAILS({share_name}) AS details",
                {
                    "share_name": bp.full_name,
                },
            )

            row = cur.fetchone()

            if row:
                for r in json_loads(row["DETAILS"]):
                    if "TARGETED WITHIN ORGANIZATION" in r["account_name"]:
                        continue

                    existing_accounts.append(AccountIdent(*r["account_name"].split(".", 2)))

        for account in bp.accounts:
            if account not in existing_accounts:
                accounts_to_add.append(account)

        for account in existing_accounts:
            if account not in bp.accounts:
                accounts_to_remove.append(account)

        if accounts_to_add:
            query = self.engine.query_builder()

            query.append(
                "ALTER SHARE {full_name:i} ADD ACCOUNTS = {accounts:i}",
                {
                    "full_name": bp.full_name,
                    "accounts": accounts_to_add,
                },
            )

            if bp.share_restrictions is not None:
                query.append_nl("SHARE_RESTRICTIONS = {share_restrictions:b}", {"share_restrictions": bp.share_restrictions})

            self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_outbound_share)

        if accounts_to_remove:
            query = self.engine.query_builder()

            query.append(
                "ALTER SHARE {full_name:i} REMOVE ACCOUNTS = {accounts:i}",
                {
                    "full_name": bp.full_name,
                    "accounts": accounts_to_remove,
                },
            )

            self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_outbound_share)

        return len(accounts_to_add) > 0 or len(accounts_to_remove) > 0

    def create_grant(self, share_name, grant: Grant):
        if grant.privilege == "USAGE" and grant.on == ObjectType.DATABASE_ROLE:
            self.engine.execute_unsafe_ddl(
                "GRANT {on:r} {name:i} TO SHARE {share_name:i}",
                {
                    "on": grant.on.singular_for_grant,
                    "name": grant.name,
                    "share_name": share_name,
                },
                condition=self.engine.settings.execute_outbound_share,
            )
        else:
            self.engine.execute_unsafe_ddl(
                "GRANT {privilege:r} ON {on:r} {name:i} TO SHARE {share_name:i}",
                {
                    "privilege": grant.privilege,
                    "on": grant.on.singular,
                    "name": grant.name,
                    "share_name": share_name,
                },
                condition=self.engine.settings.execute_outbound_share,
            )

    def drop_grant(self, share_name, grant: Grant):
        if grant.privilege == "USAGE" and grant.on == ObjectType.DATABASE_ROLE:
            self.engine.execute_unsafe_ddl(
                "REVOKE {on:r} {name:i} FROM SHARE {share_name:i}",
                {
                    "privilege": grant.privilege,
                    "on": grant.on.singular,
                    "name": grant.name,
                    "share_name": share_name,
                },
                condition=self.engine.settings.execute_outbound_share,
            )
        else:
            self.engine.execute_unsafe_ddl(
                "REVOKE {privilege:r} ON {on:r} {name:i} FROM SHARE {share_name:i}",
                {
                    "privilege": grant.privilege,
                    "on": grant.on.singular,
                    "name": grant.name,
                    "share_name": share_name,
                },
                condition=self.engine.settings.execute_outbound_share,
            )

    def get_existing_share_grants(self, share_name):
        grants = []

        cur = self.engine.execute_meta(
            "SHOW GRANTS TO SHARE {share_name:i}",
            {
                "share_name": share_name,
            },
        )

        for r in cur:
            grants.append(
                Grant(
                    privilege=r["privilege"],
                    on=ObjectType[r["granted_on"]],
                    name=build_grant_name_ident(self.config.env_prefix, r["name"], ObjectType[r["granted_on"]]),
                )
            )

        return grants

    def expand_grant_patterns_to_grants(self, grant_patterns: List[GrantPattern]):
        grants = []

        for grant_pattern in grant_patterns:
            blueprints = self.config.get_blueprints_by_type_and_pattern(grant_pattern.on.blueprint_cls, grant_pattern.pattern)

            for obj_bp in blueprints.values():
                grants.append(
                    Grant(
                        privilege=grant_pattern.privilege,
                        on=grant_pattern.on,
                        name=obj_bp.full_name,
                    ),
                )

        return grants



from snowddl.blueprint import (
    DatabaseBlueprint,
    FutureGrant,
    Grant,
    RoleBlueprint,
    SchemaBlueprint,
    build_role_ident,
)
from snowddl.resolver.abc_role_resolver import AbstractRoleResolver, ObjectType


class DatabaseOwnerRoleResolver(AbstractRoleResolver):
    def get_role_suffix(self):
        return self.config.DATABASE_ACCESS_ROLE_SUFFIX

    def get_role_type(self):
        return self.config.OWNER_ROLE_TYPE

    def get_blueprints(self):
        blueprints = []

        for database_bp in self.config.get_blueprints_by_type(DatabaseBlueprint).values():
            database_permission_model = self.config.get_permission_model(database_bp.permission_model)

            if database_permission_model.ruleset.create_database_owner_role:
                blueprints.append(self.get_blueprint_owner_role(database_bp))

        return {str(bp.full_name): bp for bp in blueprints}

    def get_blueprint_owner_role(self, database_bp: DatabaseBlueprint):
        grants = []
        account_grants = []
        future_grants = []

        database_permission_model = self.config.get_permission_model(database_bp.permission_model)

        grants.append(
            Grant(
                privilege="USAGE",
                on=ObjectType.DATABASE,
                name=database_bp.full_name,
            )
        )

        grants.append(
            Grant(
                privilege="CREATE SCHEMA",
                on=ObjectType.DATABASE,
                name=database_bp.full_name,
            )
        )

        future_grants.append(
            FutureGrant(
                privilege="OWNERSHIP",
                on_future=ObjectType.SCHEMA,
                in_parent=ObjectType.DATABASE,
                name=database_bp.full_name,
            )
        )

        # Iceberg-related grants
        if database_bp.external_volume:
            grants.append(
                Grant(
                    privilege="USAGE",
                    on=ObjectType.VOLUME,
                    name=database_bp.external_volume,
                )
            )

        if database_bp.catalog:
            grants.append(
                Grant(
                    privilege="USAGE",
                    on=ObjectType.INTEGRATION,
                    name=database_bp.catalog,
                )
            )

        # Create grants
        for model_create_grant in database_permission_model.owner_create_grants:
            future_grants.append(
                FutureGrant(
                    privilege=f"CREATE {model_create_grant.on.singular}",
                    on_future=ObjectType.SCHEMA,
                    in_parent=ObjectType.DATABASE,
                    name=database_bp.full_name,
                )
            )

        # Owner-specific grants
        for database_name_pattern in database_bp.owner_database_write:
            grants.extend(self.build_database_role_grants(database_name_pattern, self.config.WRITE_ROLE_TYPE))

        for database_name_pattern in database_bp.owner_database_read:
            grants.extend(self.build_database_role_grants(database_name_pattern, self.config.READ_ROLE_TYPE))

        for schema_name_pattern in database_bp.owner_schema_write:
            grants.extend(self.build_schema_role_grants(schema_name_pattern, self.config.WRITE_ROLE_TYPE))

        for schema_name_pattern in database_bp.owner_schema_read:
            grants.extend(self.build_schema_role_grants(schema_name_pattern, self.config.READ_ROLE_TYPE))

        for integration_name in database_bp.owner_integration_usage:
            grants.append(self.build_integration_grant(integration_name))

        for share_name in database_bp.owner_share_read:
            grants.append(self.build_share_read_grant(share_name))

        for warehouse_name in database_bp.owner_warehouse_usage:
            grants.append(self.build_warehouse_role_grant(warehouse_name, self.config.USAGE_ROLE_TYPE))

        for account_grant in database_bp.owner_account_grants:
            account_grants.append(account_grant)

        for global_role_name in database_bp.owner_global_roles:
            grants.append(self.build_global_role_grant(global_role_name))

        # Future grants on DATABASE level
        for model_future_grant in database_permission_model.owner_future_grants:
            future_grants.append(
                FutureGrant(
                    privilege=model_future_grant.privilege,
                    on_future=model_future_grant.on,
                    in_parent=ObjectType.DATABASE,
                    name=database_bp.full_name,
                )
            )

        # Future grants for each known schema on SCHEMA level
        # It is required to alleviate SCHEMA level grants overriding DATABASE level grants
        # https://docs.snowflake.com/en/sql-reference/sql/grant-privilege#future-grants-on-database-or-schema-objects
        for schema_bp in self.config.get_blueprints_by_type(SchemaBlueprint).values():
            if database_bp.full_name != schema_bp.full_name.database_full_name:
                continue

            for model_future_grant in database_permission_model.owner_future_grants:
                future_grants.append(
                    FutureGrant(
                        privilege=model_future_grant.privilege,
                        on_future=model_future_grant.on,
                        in_parent=ObjectType.SCHEMA,
                        name=schema_bp.full_name,
                    )
                )

        bp = RoleBlueprint(
            full_name=build_role_ident(
                self.config.env_prefix, database_bp.full_name.database, self.config.OWNER_ROLE_TYPE, self.get_role_suffix()
            ),
            grants=grants,
            account_grants=account_grants,
            future_grants=future_grants,
        )

        return bp



from abc import ABC, abstractmethod
from logging import getLogger, NullHandler
from pathlib import Path
from traceback import TracebackException
from typing import Callable, Dict, List, Optional, Union

from snowddl.config import SnowDDLConfig
from snowddl.blueprint import BaseDataType, NameWithType
from snowddl.parser._parsed_file import ParsedFile
from snowddl.parser._scanner import DirectoryScanner


logger = getLogger(__name__)
logger.addHandler(NullHandler())


class AbstractParser(ABC):
    def __init__(self, config: SnowDDLConfig, scanner: DirectoryScanner):
        self.config = config
        self.scanner = scanner

        self.logger = logger
        self.errors: Dict[str, Exception] = {}

        self.env_prefix = config.env_prefix

    @abstractmethod
    def load_blueprints(self):
        pass

    def get_database_names(self):
        return [database_name.upper() for database_name in self.scanner.get_database_dir_paths()]

    def get_schema_names_in_database(self, database_name):
        return [schema_name.upper() for schema_name in self.scanner.get_schema_dir_paths(database_name)]

    def parse_single_entity_file(
        self, file_key: str, json_schema: dict, callback: Callable[[ParsedFile], Union[None, Dict]] = None
    ):
        if not callback:
            callback = lambda f: f.params

        path = self.scanner.get_single_file_path(file_key)

        if path:
            try:
                file = ParsedFile(self, path, json_schema)
                return callback(file)
            except Exception as e:
                self.add_error(e, path)

        return {}

    def parse_multi_entity_file(self, file_key: str, json_schema: dict, callback: Callable[[str, Dict], None]):
        path = self.scanner.get_single_file_path(file_key)

        if path:
            try:
                file = ParsedFile(self, path, json_schema)
            except Exception as e:
                self.add_error(e, path)
                return

            for entity_name, entity_params in file.params.items():
                try:
                    callback(entity_name, entity_params)
                except Exception as e:
                    self.add_error(e, path, entity_name)

    def parse_schema_object_files(self, object_type: str, json_schema: dict, callback: Callable[[ParsedFile], None]):
        for path in self.scanner.get_schema_object_file_paths(object_type).values():
            try:
                file = ParsedFile(self, path, json_schema)
                callback(file)
            except Exception as e:
                self.add_error(e, path)

    def parse_external_file(self, path: Path, json_schema: dict, callback: Callable[[ParsedFile], Union[None, Dict]] = None):
        if not callback:
            callback = lambda f: f.params

        if path.exists():
            try:
                file = ParsedFile(self, path, json_schema)
                return callback(file)
            except Exception as e:
                self.add_error(e, path)

        return {}

    def add_error(self, exc: Exception, path: Path, entity_name: Optional[str] = None):
        traceback = "".join(TracebackException.from_exception(exc).format())

        if entity_name:
            self.logger.warning(f"Failed to parse [{entity_name}] in config file [{path}]\n{traceback}")
            error_key = f"{path}:{entity_name}"
        else:
            self.logger.warning(f"Failed to parse config file [{path}]\n{traceback}")
            error_key = str(path)

        self.errors[error_key] = exc

    def normalise_sql_text_param(self, text: str):
        return text.lstrip(" \t\n\r").rstrip(" \t\n\r;")

    def normalise_params_list(self, params):
        if params is None:
            return None

        if isinstance(params, list):
            return [p.upper() for p in params]

        raise ValueError(f"Value is neither None, nor list [{params}]")

    def normalise_params_dict(self, params):
        if params is None:
            return None

        if isinstance(params, dict):
            return {k.upper(): v for k, v in params.items()}

        raise ValueError(f"Value is neither None, nor dictionary [{params}]")

    def normalise_stage_path(self, stage_path):
        if stage_path is None:
            return None

        stage_path = str(stage_path)

        if not stage_path.startswith("/"):
            stage_path = f"/{stage_path}"

        return stage_path

    def validate_name_with_args(self, path: Path, arguments: List[NameWithType]):
        stem_name = str(path.stem)
        args_str = ",".join([a.type.base_type.name for a in arguments]).lower()

        open_pos = stem_name.find("(")
        close_pos = stem_name.find(")")

        if open_pos == -1 or close_pos == -1:
            raise ValueError(f"File [{path}] name should have list of arguments, e.g. [{stem_name}({args_str}).yaml]")

        base_name = stem_name[:open_pos]

        if stem_name[open_pos + 1 : close_pos] != args_str:
            raise ValueError(f"File [{path}] name does not match list of arguments, expected [{base_name}({args_str}).yaml]")

        # Snowflake bug: case 00444370
        # TIME and TIMESTAMP-like arguments with non-default precision are currently bugged in Snowflake
        # This check will be removed when the problem is fixed or workaround is provided
        for a in arguments:
            if (
                a.type.base_type
                in (BaseDataType.TIME, BaseDataType.TIMESTAMP_LTZ, BaseDataType.TIMESTAMP_NTZ, BaseDataType.TIMESTAMP_TZ)
                and a.type.val1 != 9
            ):
                raise ValueError(
                    f"Argument [{a.name}] with data type [{a.type.base_type.name}] must have precision of 9 (default) instead of [{a.type.val1}] due to known bug in Snowflake (case 00444370)"
                )

        return base_name



from snowddl.blueprint import AccountObjectIdent, NetworkPolicyBlueprint, SchemaObjectIdent
from snowddl.parser.abc_parser import AbstractParser


# fmt: off
network_policy_json_schema = {
    "type": "object",
    "additionalProperties": {
        "type": "object",
        "properties": {
            "allowed_network_rule_list": {
                "type": "array",
                "items": {
                    "type": "string"
                },
                "minItems": 1
            },
            "blocked_network_rule_list": {
                "type": "array",
                "items": {
                    "type": "string"
                },
                "minItems": 1
            },
            "allowed_ip_list": {
                "type": "array",
                "items": {
                    "type": "string"
                },
                "minItems": 1
            },
            "blocked_ip_list": {
                "type": "array",
                "items": {
                    "type": "string"
                },
                "minItems": 1
            },
            "comment": {
                "type": "string"
            }
        },
        "additionalProperties": False
    }
}
# fmt: on


class NetworkPolicyParser(AbstractParser):
    def load_blueprints(self):
        self.parse_multi_entity_file("network_policy", network_policy_json_schema, self.process_network_policy)

    def process_network_policy(self, policy_name, policy_params):
        policy_name_ident = AccountObjectIdent(self.env_prefix, policy_name)

        bp = NetworkPolicyBlueprint(
            full_name=policy_name_ident,
            # fmt: off
            allowed_network_rule_list=self.build_network_rule_list(policy_name_ident, policy_params.get("allowed_network_rule_list", [])),
            blocked_network_rule_list=self.build_network_rule_list(policy_name_ident, policy_params.get("blocked_network_rule_list", [])),
            # fmt: on
            allowed_ip_list=policy_params.get("allowed_ip_list", []),
            blocked_ip_list=policy_params.get("blocked_ip_list", []),
            comment=policy_params.get("comment"),
        )

        if (
            not bp.allowed_network_rule_list
            and not bp.blocked_network_rule_list
            and not bp.allowed_ip_list
            and not bp.blocked_ip_list
        ):
            raise ValueError(f"NETWORK POLICY [{bp.full_name}] must have at least one condition")

        self.config.add_blueprint(bp)

    def build_network_rule_list(self, policy_name_ident, network_rules):
        network_rule_idents = []

        for rule in network_rules:
            parts = rule.split(".")

            if len(parts) != 3:
                raise ValueError(
                    f"Network rule [{rule}] for network policy [{policy_name_ident}] should use fully-qualified identifier <database>.<schema>.<name>"
                )

            network_rule_idents.append(SchemaObjectIdent(self.env_prefix, *parts))

        return network_rule_idents



from snowddl.blueprint import AccountObjectIdent, NetworkPolicyBlueprint
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType

from json import loads


class NetworkPolicyResolver(AbstractResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.NETWORK_POLICY

    def get_existing_objects(self):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW NETWORK POLICIES LIKE {env_prefix:ls}",
            {
                "env_prefix": self.config.env_prefix,
            },
        )

        for r in cur:
            existing_objects[r["name"]] = {
                "name": r["name"],
                "comment": r["comment"] if r["comment"] else None,
            }

        # Ownership is not available in SHOW NETWORK POLICIES
        # But it can be derived from grants
        for name, owner in self.engine.executor.map(self.get_owner_from_grant, existing_objects.keys()):
            if owner != self.engine.context.current_role:
                del existing_objects[name]

        return existing_objects

    def get_owner_from_grant(self, name):
        cur = self.engine.execute_meta(
            "SHOW GRANTS ON NETWORK POLICY {name:i}",
            {
                "name": name,
            },
        )

        for r in cur:
            # Assumption: OWNERSHIP grant always exist, exactly 1 row
            if r["privilege"] == "OWNERSHIP":
                return name, r["grantee_name"]

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(NetworkPolicyBlueprint)

    def create_object(self, bp: NetworkPolicyBlueprint):
        self._create_policy(bp)
        self._apply_policy_refs(bp, skip_existing=True)

        return ResolveResult.CREATE

    def compare_object(self, bp: NetworkPolicyBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        cur = self.engine.execute_meta(
            "DESC NETWORK POLICY {name:i}",
            {
                "name": bp.full_name,
            },
        )

        existing_allowed_network_rule_list = []
        existing_blocked_network_rule_list = []
        existing_allowed_ip_list = []
        existing_blocked_ip_list = []

        for r in cur:
            if r["name"] == "ALLOWED_NETWORK_RULE_LIST":
                existing_allowed_network_rule_list = [v["fullyQualifiedRuleName"] for v in loads(r["value"])]
            elif r["name"] == "BLOCKED_NETWORK_RULE_LIST":
                existing_blocked_network_rule_list = [v["fullyQualifiedRuleName"] for v in loads(r["value"])]
            elif r["name"] == "ALLOWED_IP_LIST":
                existing_allowed_ip_list = str(r["value"]).split(",")
            elif r["name"] == "BLOCKED_IP_LIST":
                existing_blocked_ip_list = str(r["value"]).split(",")

        if sorted(map(str, bp.allowed_network_rule_list)) != sorted(existing_allowed_network_rule_list):
            if bp.allowed_network_rule_list:
                self.engine.execute_unsafe_ddl(
                    "ALTER NETWORK POLICY {name:i} SET ALLOWED_NETWORK_RULE_LIST = ({allowed_network_rule_list:i})",
                    {
                        "name": bp.full_name,
                        "allowed_network_rule_list": bp.allowed_network_rule_list,
                    },
                    condition=self.engine.settings.execute_network_policy,
                )
            else:
                self.engine.execute_unsafe_ddl(
                    "ALTER NETWORK POLICY {name:i} UNSET ALLOWED_NETWORK_RULE_LIST",
                    {
                        "name": bp.full_name,
                    },
                    condition=self.engine.settings.execute_network_policy,
                )

            result = ResolveResult.ALTER

        if sorted(map(str, bp.blocked_network_rule_list)) != sorted(existing_blocked_network_rule_list):
            if bp.blocked_network_rule_list:
                self.engine.execute_unsafe_ddl(
                    "ALTER NETWORK POLICY {name:i} SET BLOCKED_NETWORK_RULE_LIST = ({blocked_network_rule_list:i})",
                    {
                        "name": bp.full_name,
                        "blocked_network_rule_list": bp.blocked_network_rule_list,
                    },
                    condition=self.engine.settings.execute_network_policy,
                )
            else:
                self.engine.execute_unsafe_ddl(
                    "ALTER NETWORK POLICY {name:i} UNSET BLOCKED_NETWORK_RULE_LIST",
                    {
                        "name": bp.full_name,
                    },
                    condition=self.engine.settings.execute_network_policy,
                )

            result = ResolveResult.ALTER

        if sorted(bp.allowed_ip_list) != sorted(existing_allowed_ip_list):
            if bp.allowed_ip_list:
                self.engine.execute_unsafe_ddl(
                    "ALTER NETWORK POLICY {name:i} SET ALLOWED_IP_LIST = ({blocked_ip_list})",
                    {
                        "name": bp.full_name,
                        "blocked_ip_list": bp.allowed_ip_list,
                    },
                    condition=self.engine.settings.execute_network_policy,
                )
            else:
                self.engine.execute_unsafe_ddl(
                    "ALTER NETWORK POLICY {name:i} UNSET ALLOWED_IP_LIST",
                    {
                        "name": bp.full_name,
                    },
                    condition=self.engine.settings.execute_network_policy,
                )

            result = ResolveResult.ALTER

        if sorted(bp.blocked_ip_list) != sorted(existing_blocked_ip_list):
            if bp.blocked_ip_list:
                self.engine.execute_unsafe_ddl(
                    "ALTER NETWORK POLICY {name:i} SET BLOCKED_IP_LIST = ({blocked_ip_list})",
                    {
                        "name": bp.full_name,
                        "blocked_ip_list": bp.blocked_ip_list,
                    },
                    condition=self.engine.settings.execute_network_policy,
                )
            else:
                self.engine.execute_unsafe_ddl(
                    "ALTER NETWORK POLICY {name:i} UNSET BLOCKED_IP_LIST",
                    {
                        "name": bp.full_name,
                    },
                    condition=self.engine.settings.execute_network_policy,
                )

            result = ResolveResult.ALTER

        if bp.comment != row["comment"]:
            self.engine.execute_unsafe_ddl(
                "ALTER NETWORK POLICY {name:i} SET COMMENT = {comment}",
                {
                    "name": bp.full_name,
                    "comment": bp.comment if bp.comment else None,
                },
                condition=self.engine.settings.execute_network_policy,
            )

            result = ResolveResult.ALTER

        if self._apply_policy_refs(bp):
            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        policy_name = AccountObjectIdent("", row["name"])

        self._drop_policy_refs(policy_name)
        self._drop_policy(policy_name)

        return ResolveResult.DROP

    def _create_policy(self, bp: NetworkPolicyBlueprint):
        query = self.engine.query_builder()

        query.append(
            "CREATE NETWORK POLICY {name:i}",
            {
                "name": bp.full_name,
            },
        )

        if bp.allowed_network_rule_list:
            query.append_nl(
                "ALLOWED_NETWORK_RULE_LIST = ({allowed_network_rule_list:i})",
                {
                    "allowed_network_rule_list": bp.allowed_network_rule_list,
                },
            )

        if bp.blocked_network_rule_list:
            query.append_nl(
                "BLOCKED_NETWORK_RULE_LIST = ({blocked_network_rule_list:i})",
                {
                    "blocked_network_rule_list": bp.blocked_network_rule_list,
                },
            )

        if bp.allowed_ip_list:
            query.append_nl(
                "ALLOWED_IP_LIST = ({allowed_ip_list})",
                {
                    "allowed_ip_list": bp.allowed_ip_list,
                },
            )

        if bp.blocked_ip_list:
            query.append_nl(
                "BLOCKED_IP_LIST = ({blocked_ip_list})",
                {
                    "blocked_ip_list": bp.blocked_ip_list,
                },
            )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_network_policy)

    def _drop_policy(self, policy_name: AccountObjectIdent):
        self.engine.execute_unsafe_ddl(
            "DROP NETWORK POLICY {full_name:i}",
            {"full_name": policy_name},
            condition=self.engine.settings.execute_network_policy,
        )

    def _apply_policy_refs(self, bp: NetworkPolicyBlueprint, skip_existing=False):
        existing_policy_refs = {} if skip_existing else self._get_existing_policy_refs(bp.full_name)
        applied_change = False

        for ref in bp.references:
            if ref.object_type == ObjectType.ACCOUNT:
                ref_key = ref.object_type.name
            else:
                ref_key = f"{ref.object_type.name}|{ref.object_name}"

            # Policy was applied before
            if ref_key in existing_policy_refs:
                del existing_policy_refs[ref_key]
                continue

            if ref.object_type == ObjectType.ACCOUNT:
                # Apply new policy for ACCOUNT
                self.engine.execute_unsafe_ddl(
                    "ALTER ACCOUNT SET NETWORK_POLICY = {policy_name:i}",
                    {
                        "policy_name": bp.full_name,
                    },
                    condition=self.engine.settings.execute_network_policy and self.engine.settings.execute_account_level_policy,
                )
            else:
                # Apply new policy for USER (and other object types in future?)
                self.engine.execute_unsafe_ddl(
                    "ALTER {object_type:r} {object_name:i} SET NETWORK_POLICY = {policy_name:i}",
                    {
                        "object_type": ref.object_type.name,
                        "object_name": ref.object_name,
                        "policy_name": bp.full_name,
                    },
                    condition=self.engine.settings.execute_network_policy,
                )

            applied_change = True

        # Remove remaining policy references which no longer exist in blueprint
        for existing_ref in existing_policy_refs.values():
            if existing_ref["object_type"] == ObjectType.ACCOUNT.name:
                self.engine.execute_unsafe_ddl(
                    "ALTER ACCOUNT UNSET NETWORK_POLICY",
                    condition=self.engine.settings.execute_network_policy and self.engine.settings.execute_account_level_policy,
                )
            else:
                self.engine.execute_unsafe_ddl(
                    "ALTER {object_type:r} {object_name:i} UNSET NETWORK_POLICY",
                    {
                        "object_type": existing_ref["object_type"],
                        "object_name": existing_ref["name"],
                    },
                    condition=self.engine.settings.execute_network_policy,
                )

            applied_change = True

        return applied_change

    def _drop_policy_refs(self, policy_name: AccountObjectIdent):
        existing_policy_refs = self._get_existing_policy_refs(policy_name)

        for existing_ref in existing_policy_refs.values():
            if existing_ref["object_type"] == ObjectType.ACCOUNT.name:
                self.engine.execute_unsafe_ddl(
                    "ALTER ACCOUNT UNSET NETWORK_POLICY",
                    condition=self.engine.settings.execute_network_policy and self.engine.settings.execute_account_level_policy,
                )
            else:
                self.engine.execute_unsafe_ddl(
                    "ALTER {object_type:r} {object_name:i} UNSET NETWORK_POLICY",
                    {
                        "object_type": existing_ref["object_type"],
                        "object_name": existing_ref["name"],
                    },
                    condition=self.engine.settings.execute_network_policy,
                )

    def _get_existing_policy_refs(self, policy_name: AccountObjectIdent):
        existing_policy_refs = {}

        cur = self.engine.execute_meta(
            "SELECT * FROM TABLE(snowflake.information_schema.policy_references(policy_name => {policy_name}, policy_kind => {policy_kind}))",
            {
                "policy_name": policy_name,
                "policy_kind": self.object_type.name,
            },
        )

        for r in cur:
            if r["REF_ENTITY_DOMAIN"] == ObjectType.ACCOUNT.name:
                ref_key = r["REF_ENTITY_DOMAIN"]
            else:
                ref_key = f"{r['REF_ENTITY_DOMAIN']}|{r['REF_ENTITY_NAME']}"

            existing_policy_refs[ref_key] = {
                "object_type": r["REF_ENTITY_DOMAIN"],
                "name": r["REF_ENTITY_NAME"],
            }

        return existing_policy_refs



from snowddl.blueprint import (
    ExternalFunctionBlueprint,
    Ident,
    SchemaObjectIdentWithArgs,
    NameWithType,
    DataType,
    build_schema_object_ident,
)
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
external_function_json_schema = {
    "type": "object",
    "properties": {
        "arguments": {
            "type": "object",
            "additionalProperties": {
                "type": "string"
            }
        },
        "returns": {
            "type": "string"
        },
        "api_integration": {
            "type": "string"
        },
        "url": {
            "type": "string",
        },
        "is_secure": {
            "type": "boolean"
        },
        "is_strict": {
            "type": "boolean"
        },
        "is_immutable": {
            "type": "boolean"
        },
        "headers": {
            "type": "object",
            "additionalProperties": {
                "type": "string"
            }
        },
        "context_headers": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "max_batch_rows": {
            "type": "integer"
        },
        "compression": {
            "type": "string"
        },
        "request_translator": {
            "type": "string"
        },
        "response_translator": {
            "type": "string"
        },
        "comment": {
            "type": "string"
        }
    },
    "required": ["api_integration", "url"],
    "additionalProperties": False
}
# fmt: on


class ExternalFunctionParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("external_function", external_function_json_schema, self.process_function)

    def process_function(self, f: ParsedFile):
        arguments = [NameWithType(name=Ident(k), type=DataType(t)) for k, t in f.params.get("arguments", {}).items()]
        returns = DataType(f.params["returns"])

        base_name = self.validate_name_with_args(f.path, arguments)

        bp = ExternalFunctionBlueprint(
            full_name=SchemaObjectIdentWithArgs(
                self.env_prefix, f.database, f.schema, base_name, [a.type.base_type for a in arguments]
            ),
            arguments=arguments,
            returns=returns,
            api_integration=Ident(f.params.get("api_integration")),
            url=f.params["url"],
            is_secure=f.params.get("is_secure", False),
            is_strict=f.params.get("is_strict", False),
            is_immutable=f.params.get("is_immutable", False),
            headers=f.params.get("headers"),
            context_headers=[Ident(h) for h in f.params.get("context_headers")] if f.params.get("context_headers") else None,
            max_batch_rows=f.params.get("max_batch_rows"),
            compression=f.params.get("compression"),
            request_translator=build_schema_object_ident(self.env_prefix, f.params["request_translator"], f.database, f.schema)
            if f.params.get("request_translator")
            else None,
            response_translator=build_schema_object_ident(self.env_prefix, f.params["response_translator"], f.database, f.schema)
            if f.params.get("response_translator")
            else None,
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)



from functools import partial

from snowddl.blueprint import (
    UserBlueprint,
    AccountObjectIdent,
    AuthenticationPolicyBlueprint,
    AuthenticationPolicyReference,
    NetworkPolicyBlueprint,
    NetworkPolicyReference,
    ObjectType,
    SchemaObjectIdent,
    build_role_ident,
    build_default_namespace_ident,
)
from snowddl.parser.abc_parser import AbstractParser
from snowddl.parser.business_role import business_role_json_schema


# fmt: off
user_json_schema = {
    "type": "object",
    "additionalProperties": {
        "properties": {
            "login_name": {
                "type": "string"
            },
            "display_name": {
                "type": "string"
            },
            "first_name": {
                "type": "string"
            },
            "last_name": {
                "type": "string"
            },
            "email": {
                "type": "string"
            },
            "disabled": {
                "type": "boolean"
            },
            "password": {
                "type": "string"
            },
            "rsa_public_key": {
                "type": "string"
            },
            "rsa_public_key_2": {
                "type": "string"
            },
            "type": {
                "type": "string",
            },
            "default_warehouse": {
                "type": "string"
            },
            "default_namespace": {
                "type": "string"
            },
            "session_params": {
                "type": "object",
                "additionalProperties": {
                    "type": ["boolean", "number", "string"]
                }
            },
            "business_roles": {
                "type": "array",
                "items": {
                    "type": "string"
                },
                "minItems": 1
            },
            "comment": {
                "type": "string"
            },
            "authentication_policy": {
                "type": "string",
            },
            "network_policy": {
                "type": "string",
            },
            "workload_identity": {
                "type": "object",
                "additionalProperties": {
                    "type": ["array", "boolean", "number", "string"]
                }
            },
        },
        "additionalProperties": False
    }
}
# fmt: on


class UserParser(AbstractParser):
    def load_blueprints(self):
        default_wh_map = self.get_default_warehouse_map()

        self.parse_multi_entity_file("user", user_json_schema, partial(self.process_user, default_wh_map=default_wh_map))

    def process_user(self, user_name, user_params, default_wh_map):
        business_roles = []
        default_warehouse = user_params.get("default_warehouse")

        for business_role_name in user_params.get("business_roles", []):
            business_roles.append(build_role_ident(self.env_prefix, business_role_name, self.config.BUSINESS_ROLE_SUFFIX))

            if default_warehouse is None:
                default_warehouse = default_wh_map.get(business_role_name)

        full_user_name = AccountObjectIdent(self.env_prefix, user_name)

        if user_params.get("login_name"):
            # Login name is string, not identifier (!), special characters like '@#!' are permitted
            # But it still requires env_prefix due to unique constraint
            login_name = f"{self.config.env_prefix}{user_params.get('login_name')}".upper()
        else:
            login_name = str(full_user_name)

        if user_params.get("display_name"):
            display_name = str(user_params.get("display_name")).upper()
        else:
            display_name = str(full_user_name)

        # fmt: off
        bp = UserBlueprint(
            full_name=full_user_name,
            login_name=login_name,
            display_name=display_name,
            first_name=user_params.get("first_name"),
            last_name=user_params.get("last_name"),
            email=user_params.get("email"),
            disabled=user_params.get("disabled", False),
            password=user_params.get("password"),
            rsa_public_key=user_params.get("rsa_public_key").replace(" ", "") if user_params.get("rsa_public_key") else None,
            rsa_public_key_2=user_params.get("rsa_public_key_2").replace(" ", "") if user_params.get("rsa_public_key_2") else None,
            type=str(user_params.get("type")).upper() if user_params.get("type") else None,
            default_warehouse=AccountObjectIdent(self.env_prefix, default_warehouse) if default_warehouse else None,
            default_namespace=build_default_namespace_ident(self.env_prefix, user_params.get("default_namespace")) if user_params.get("default_namespace") else None,
            session_params=self.normalise_params_dict(user_params.get("session_params", {})),
            workload_identity=self.normalise_params_dict(user_params.get("workload_identity")),
            business_roles=business_roles,
            comment=user_params.get("comment"),
        )
        # fmt: on

        self.config.add_blueprint(bp)

        # Authentication policy
        if user_params.get("authentication_policy"):
            policy_name_parts = user_params.get("authentication_policy").split(".")

            if len(policy_name_parts) != 3:
                raise ValueError(
                    f"Authentication policy [{user_params.get('authentication_policy')}] should use fully-qualified identifier <database>.<schema>.<name> for user [{full_user_name}]"
                )

            policy_name = SchemaObjectIdent(self.env_prefix, *policy_name_parts)

            ref = AuthenticationPolicyReference(
                object_type=ObjectType.USER,
                object_name=full_user_name,
            )

            self.config.add_policy_reference(AuthenticationPolicyBlueprint, policy_name, ref)

        # Network policy
        if user_params.get("network_policy"):
            policy_name = AccountObjectIdent(self.env_prefix, user_params.get("network_policy"))

            ref = NetworkPolicyReference(
                object_type=ObjectType.USER,
                object_name=full_user_name,
            )

            self.config.add_policy_reference(NetworkPolicyBlueprint, policy_name, ref)

        if "NETWORK_POLICY" in bp.session_params:
            raise ValueError(
                "NETWORK_POLICY in session_params of USER is no longer supported. Please use dedicated [network_policy] parameter instead. Read more: https://docs.snowddl.com/breaking-changes-log/0.33.0-october-2024"
            )

    def get_default_warehouse_map(self):
        default_warehouse_map = {}

        business_role_config = self.parse_single_entity_file("business_role", business_role_json_schema)

        for business_role_name, business_role in business_role_config.items():
            if "warehouse_usage" not in business_role:
                continue

            default_warehouse_map[business_role_name] = business_role["warehouse_usage"][0]

        return default_warehouse_map



from snowddl.blueprint import RoleBlueprint, WarehouseBlueprint, Grant, build_role_ident
from snowddl.resolver.abc_role_resolver import AbstractRoleResolver, ObjectType


class WarehouseUsageRoleResolver(AbstractRoleResolver):
    def get_role_suffix(self):
        return self.config.WAREHOUSE_ACCESS_ROLE_SUFFIX

    def get_role_type(self):
        return self.config.USAGE_ROLE_TYPE

    def get_blueprints(self):
        blueprints = []

        for warehouse in self.config.get_blueprints_by_type(WarehouseBlueprint).values():
            blueprints.append(self.get_blueprint_usage_role(warehouse))

        return {str(bp.full_name): bp for bp in blueprints}

    def get_blueprint_usage_role(self, warehouse: WarehouseBlueprint):
        grants = []

        grants.append(
            Grant(
                privilege="USAGE",
                on=ObjectType.WAREHOUSE,
                name=warehouse.full_name,
            )
        )

        grants.append(
            Grant(
                privilege="OPERATE",
                on=ObjectType.WAREHOUSE,
                name=warehouse.full_name,
            )
        )

        bp = RoleBlueprint(
            full_name=build_role_ident(
                self.config.env_prefix, warehouse.full_name.name, self.config.USAGE_ROLE_TYPE, self.get_role_suffix()
            ),
            grants=grants,
            future_grants=[],
        )

        return bp



from typing import Union


def coalesce(val, default=None):
    return default if val is None else val


def compare_dynamic_param_value(bp_value: Union[bool, int, float, str], existing_value: str) -> bool:
    if isinstance(bp_value, bool):
        if bp_value is True and existing_value == "true":
            return True

        if bp_value is False and existing_value == "false":
            return True

    elif isinstance(bp_value, int):
        if bp_value == int(existing_value):
            return True

    elif isinstance(bp_value, float):
        if bp_value == float(existing_value):
            return True

    else:
        if str(bp_value) == existing_value:
            return True

    return False


def dtypes_from_arguments(arguments: str) -> str:
    all_dtypes = []

    start_dtypes_idx = arguments.index("(")
    finish_dtypes_idx = arguments.index(") RETURN ")

    for dtype_part in split_by_comma_outside_parentheses(arguments[start_dtypes_idx + 1 : finish_dtypes_idx]):
        # Remove optional DEFAULT prefix from the beginning
        dtype_part = dtype_part.removeprefix("DEFAULT ")

        # Remove optional data type size introduced in bundle 2025_03
        # https://docs.snowflake.com/en/release-notes/bcr-bundles/2025_03/bcr-1944
        dtype_size_start_idx = dtype_part.find("(")

        if dtype_size_start_idx > -1:
            dtype_part = dtype_part[:dtype_size_start_idx]

        all_dtypes.append(dtype_part)

    return ",".join(all_dtypes)


def split_by_comma_outside_parentheses(s: str):
    parts = []
    current = []
    depth = 0

    for char in s:
        if char == "," and depth == 0:
            parts.append("".join(current).strip())
            current = []

        else:
            if char == "(":
                depth += 1
            elif char == ")":
                depth -= 1

            current.append(char)

    if current:
        parts.append("".join(current).strip())

    return parts



from snowddl.blueprint import AuthenticationPolicyBlueprint, SchemaObjectIdent
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
authentication_policy_json_schema = {
    "type": "object",
    "properties": {
        "authentication_methods": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "mfa_authentication_methods": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "mfa_enrollment": {
            "type": "string"
        },
        "mfa_policy": {
            "type": "object",
            "additionalProperties": {
                "type": ["array", "boolean", "number", "string"]
            }
        },
        "client_types": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "client_policy": {
            "type": "object",
            "additionalProperties": {
                "type": "object",
                "additionalProperties": {
                    "type": ["array", "boolean", "number", "string"]
                }
            }
        },
        "security_integrations": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "pat_policy": {
            "type": "object",
            "additionalProperties": {
                "type": ["array", "boolean", "number", "string"]
            }
        },
        "workload_identity_policy": {
            "type": "object",
            "additionalProperties": {
                "type": ["array", "boolean", "number", "string"]
            }
        },
        "comment": {
            "type": "string"
        }
    },
    "additionalProperties": False,
}
# fmt: on


class AuthenticationPolicyParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files(
            "authentication_policy", authentication_policy_json_schema, self.process_authentication_policy
        )

    def process_authentication_policy(self, f: ParsedFile):
        bp = AuthenticationPolicyBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            authentication_methods=self.normalise_params_list(f.params.get("authentication_methods")),
            mfa_authentication_methods=self.normalise_params_list(f.params.get("mfa_authentication_methods")),
            mfa_enrollment=f.params.get("mfa_enrollment").upper() if f.params.get("mfa_enrollment") else None,
            mfa_policy=self.normalise_params_dict(f.params.get("mfa_policy")),
            client_types=self.normalise_params_list(f.params.get("client_types")),
            client_policy=self._normalise_client_policy(f.params.get("client_policy")),
            security_integrations=self.normalise_params_list(f.params.get("security_integrations")),
            pat_policy=self.normalise_params_dict(f.params.get("pat_policy")),
            workload_identity_policy=self.normalise_params_dict(f.params.get("workload_identity_policy")),
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)

    def _normalise_client_policy(self, client_policy):
        if client_policy is None:
            return None

        result = {}

        for k, v in client_policy.items():
            result[k.upper()] = self.normalise_params_dict(v)

        return result



from typing import Optional, List

from snowddl.blueprint import DatabaseIdent, ObjectType, Ident
from snowddl.model import BaseModelWithConfig


class SnowDDLSettings(BaseModelWithConfig):
    env_admin_role: Optional[Ident] = None
    execute_safe_ddl: bool = False
    execute_unsafe_ddl: bool = False
    execute_replace_table: bool = False
    execute_account_level_policy: bool = False
    execute_aggregation_policy: bool = False
    execute_authentication_policy: bool = False
    execute_masking_policy: bool = False
    execute_projection_policy: bool = False
    execute_row_access_policy: bool = False
    execute_account_params: bool = False
    execute_network_policy: bool = False
    execute_resource_monitor: bool = False
    execute_outbound_share: bool = False
    refresh_user_passwords: bool = False
    refresh_workload_identity: bool = False
    refresh_future_grants: bool = False
    refresh_stage_encryption: bool = False
    refresh_secrets: bool = False
    clone_table: bool = False
    clone_source_env_prefix: str = ""
    exclude_object_types: List[ObjectType] = []
    include_object_types: List[ObjectType] = []
    include_databases: List[DatabaseIdent] = []
    ignore_ownership: bool = False
    max_workers: int = 32

    # Options specific for snowddl-convert
    convert_function_body_to_file: bool = False
    convert_view_text_to_file: bool = False



from json import loads as json_loads

from snowddl.blueprint import DynamicTableBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class DynamicTableResolver(AbstractSchemaObjectResolver):
    skip_on_empty_blueprints = True

    unit_to_seconds_multiplier = {
        "second": 1,
        "seconds": 1,
        "minute": 60,
        "minutes": 60,
        "hour": 3600,
        "hours": 3600,
        "day": 86400,
        "days": 86400,
    }

    def get_object_type(self) -> ObjectType:
        return ObjectType.DYNAMIC_TABLE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW AS RESOURCE DYNAMIC TABLES IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            r = json_loads(r["As Resource"])

            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                "is_transient": r["kind"] == "TRANSIENT",
                "retention_time": r["data_retention_time_in_days"],
                "columns": r["columns"],
                "text": r["query"].rstrip(";"),
                "cluster_by": r["cluster_by"],
                "target_lag": r["target_lag"],
                "refresh_mode": r["refresh_mode"],
                "warehouse": r["warehouse"],
                "comment": r["comment"],
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(DynamicTableBlueprint)

    def create_object(self, bp: DynamicTableBlueprint):
        query = self.engine.query_builder()
        query.append("CREATE OR REPLACE")

        if bp.is_transient:
            query.append("TRANSIENT")

        query.append(
            "DYNAMIC TABLE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        query.append(self._build_common_dynamic_table_sql(bp))

        self.engine.execute_safe_ddl(query)

        return ResolveResult.CREATE

    def compare_object(self, bp: DynamicTableBlueprint, row: dict):
        result = ResolveResult.NOCHANGE
        replace_reasons = []

        if bp.columns and [str(c.name) for c in bp.columns] != [str(c["name"]) for c in row["columns"]]:
            replace_reasons.append("Column definition was changed")

        if bp.text != row["text"]:
            replace_reasons.append("SQL text was changed")

        if bp.is_transient is True and row["is_transient"] is False:
            replace_reasons.append("Dynamic table type was changed to TRANSIENT")
        elif bp.is_transient is False and row["is_transient"] is True:
            replace_reasons.append("Dynamic table type was changed to PERMANENT")

        if bp.refresh_mode and bp.refresh_mode != "AUTO" and bp.refresh_mode != row["refresh_mode"]:
            replace_reasons.append(f"Refresh mode was changed to {bp.refresh_mode}")

        if replace_reasons:
            query = self.engine.query_builder()
            query.append("CREATE OR REPLACE")

            if bp.is_transient:
                query.append("TRANSIENT")

            query.append(
                "DYNAMIC TABLE {full_name:i}",
                {
                    "full_name": bp.full_name,
                },
            )

            query.append(self._build_common_dynamic_table_sql(bp))
            self.engine.execute_unsafe_ddl("\n".join(f"-- {r}" for r in replace_reasons) + "\n" + str(query))

            return ResolveResult.REPLACE

        if not self._compare_target_lag(bp, row):
            self.engine.execute_safe_ddl(
                "ALTER DYNAMIC TABLE {full_name:i} SET TARGET_LAG = {target_lag}",
                {
                    "full_name": bp.full_name,
                    "target_lag": bp.target_lag,
                },
            )

            result = ResolveResult.ALTER

        if str(bp.warehouse) != row["warehouse"]:
            self.engine.execute_safe_ddl(
                "ALTER DYNAMIC TABLE {full_name:i} SET WAREHOUSE = {warehouse:i}",
                {
                    "full_name": bp.full_name,
                    "warehouse": bp.warehouse,
                },
            )

            result = ResolveResult.ALTER

        if not self._compare_cluster_by(bp, row):
            if bp.cluster_by:
                self.engine.execute_unsafe_ddl(
                    "ALTER DYNAMIC TABLE {full_name:i} CLUSTER BY ({cluster_by:r})",
                    {
                        "full_name": bp.full_name,
                        "cluster_by": bp.cluster_by,
                    },
                )
            else:
                self.engine.execute_unsafe_ddl(
                    "ALTER DYNAMIC TABLE {full_name:i} DROP CLUSTERING KEY",
                    {"full_name": bp.full_name},
                )

            result = ResolveResult.ALTER

        if bp.retention_time is not None and bp.retention_time != row["retention_time"]:
            self.engine.execute_unsafe_ddl(
                "ALTER DYNAMIC TABLE {full_name:i} SET DATA_RETENTION_TIME_IN_DAYS = {retention_time:d}",
                {
                    "full_name": bp.full_name,
                    "retention_time": bp.retention_time,
                },
            )

            result = ResolveResult.ALTER

        if bp.comment != row["comment"]:
            self.engine.execute_safe_ddl(
                "ALTER DYNAMIC TABLE {full_name:i} SET COMMENT = {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": bp.comment,
                },
            )

            result = ResolveResult.ALTER

        for idx, c in enumerate(row["columns"]):
            bp_col_comment = bp.columns[idx].comment if bp.columns else None

            if bp_col_comment != c["comment"]:
                if bp_col_comment:
                    self.engine.execute_safe_ddl(
                        "ALTER DYNAMIC TABLE {full_name:i} MODIFY COLUMN {column_name:i} COMMENT {comment}",
                        {
                            "full_name": bp.full_name,
                            "column_name": c["name"],
                            "comment": bp_col_comment,
                        },
                    )
                else:
                    self.engine.execute_safe_ddl(
                        "ALTER DYNAMIC TABLE {full_name:i} MODIFY COLUMN {column_name:i} UNSET COMMENT",
                        {
                            "full_name": bp.full_name,
                            "column_name": c["name"],
                        },
                    )

                result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP DYNAMIC TABLE {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _build_common_dynamic_table_sql(self, bp: DynamicTableBlueprint):
        query = self.engine.query_builder()

        if bp.columns:
            query.append_nl("(")

            for idx, c in enumerate(bp.columns):
                query.append_nl(
                    "    {comma:r}{col_name:i}",
                    {
                        "comma": "  " if idx == 0 else ", ",
                        "col_name": c.name,
                    },
                )

                if c.comment:
                    query.append(
                        "COMMENT {col_comment}",
                        {
                            "col_comment": c.comment,
                        },
                    )

            query.append_nl(")")

        query.append_nl(
            "TARGET_LAG = {target_lag}",
            {
                "target_lag": bp.target_lag,
            },
        )

        query.append_nl(
            "WAREHOUSE = {warehouse:i}",
            {
                "warehouse": bp.warehouse,
            },
        )

        if bp.refresh_mode:
            query.append_nl(
                "REFRESH_MODE = {refresh_mode}",
                {
                    "refresh_mode": bp.refresh_mode,
                },
            )

        if bp.initialize:
            query.append_nl(
                "INITIALIZE = {initialize}",
                {
                    "initialize": bp.initialize,
                },
            )

        if bp.cluster_by:
            query.append_nl(
                "CLUSTER BY ({cluster_by:r})",
                {
                    "cluster_by": bp.cluster_by,
                },
            )

        if bp.retention_time is not None:
            query.append_nl(
                "DATA_RETENTION_TIME_IN_DAYS = {retention_time:d}",
                {
                    "retention_time": bp.retention_time,
                },
            )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        query.append_nl("AS")
        query.append_nl(bp.text)

        return query

    def _compare_cluster_by(self, bp: DynamicTableBlueprint, row: dict):
        bp_cluster_by = ", ".join(bp.cluster_by).upper() if bp.cluster_by else None
        snow_cluster_by = ", ".join(row["cluster_by"]).upper() if row["cluster_by"] else None

        return bp_cluster_by == snow_cluster_by

    def _compare_target_lag(self, bp: DynamicTableBlueprint, row: dict):
        if bp.target_lag == "DOWNSTREAM":
            return row["target_lag"]["type"] == "DOWNSTREAM"

        num, _, unit = bp.target_lag.partition(" ")

        num = int(num)
        unit = unit.lower()

        num_in_seconds = num * self.unit_to_seconds_multiplier[unit]

        if row["target_lag"]["type"] == "USER_DEFINED" and row["target_lag"]["seconds"] == num_in_seconds:
            return True

        return False



from snowddl.blueprint import StageBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType
from snowddl.resolver._utils import coalesce, compare_dynamic_param_value


class StageResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.STAGE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW STAGES IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            if "TEMPORARY" in r["type"]:
                # Skip TEMPORARY stages
                continue

            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "url": r["url"] if r["url"] else None,
                "storage_integration": r["storage_integration"] if r["storage_integration"] else None,
                "owner": r["owner"],
                "type": r["type"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(StageBlueprint)

    def create_object(self, bp: StageBlueprint):
        query = self.engine.query_builder()
        query.append(
            "CREATE STAGE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        query.append_nl(self._build_create_stage_sql(bp))
        self.engine.execute_safe_ddl(query)

        return ResolveResult.CREATE

    def compare_object(self, bp: StageBlueprint, row: dict):
        if self._is_replace_required(bp, row):
            query = self.engine.query_builder()
            query.append(
                "CREATE OR REPLACE STAGE {full_name:i}",
                {
                    "full_name": bp.full_name,
                },
            )

            query.append_nl(self._build_create_stage_sql(bp))
            self.engine.execute_unsafe_ddl(query)

            return ResolveResult.REPLACE

        # Check specific properties for potential ALTER
        existing_properties = self._get_existing_stage_properties(bp)
        result = ResolveResult.NOCHANGE

        if self._compare_url(bp, row):
            result = ResolveResult.ALTER

        if self._compare_storage_integration(bp, row):
            result = ResolveResult.ALTER

        if self._compare_directory(bp, existing_properties):
            result = ResolveResult.ALTER

        if self._compare_file_format(bp, existing_properties):
            result = ResolveResult.ALTER

        if self._compare_copy_options(bp, existing_properties):
            result = ResolveResult.ALTER

        if self._compare_comment(bp, row):
            result = ResolveResult.ALTER

        if self.engine.settings.refresh_stage_encryption and bp.url and self._refresh_encryption(bp):
            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP STAGE {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _is_replace_required(self, bp: StageBlueprint, row):
        # Change of stage type from INTERNAL to EXTERNAL or vice versa
        if (bp.url and not row["url"]) or (not bp.url and row["url"]):
            return True

        if bp.url is None:
            encryption_type = coalesce(bp.encryption, {}).get("TYPE", "SNOWFLAKE_FULL").upper()

            if row["type"] == "INTERNAL" and encryption_type != "SNOWFLAKE_FULL":
                return True

            if row["type"] == "INTERNAL NO CSE" and encryption_type != "SNOWFLAKE_SSE":
                return True

        return False

    def _compare_url(self, bp: StageBlueprint, row):
        if not ((bp.url is None and row["url"] is None) or str(bp.url) == row["url"]):
            query = self.engine.query_builder()
            query.append(
                "ALTER STAGE {full_name:i} SET",
                {
                    "full_name": bp.full_name,
                },
            )

            query.append_nl(
                "URL = {url:i}",
                {
                    "url": bp.url,
                },
            )

            self.engine.execute_safe_ddl(query)

            return True

        return False

    def _compare_storage_integration(self, bp: StageBlueprint, row):
        if not (
            (bp.storage_integration is None and row["storage_integration"] is None)
            or str(bp.storage_integration) == row["storage_integration"]
        ):
            query = self.engine.query_builder()
            query.append(
                "ALTER STAGE {full_name:i} SET",
                {
                    "full_name": bp.full_name,
                },
            )

            query.append_nl(
                "STORAGE_INTEGRATION = {storage_integration:i}",
                {
                    "storage_integration": bp.storage_integration,
                },
            )

            self.engine.execute_safe_ddl(query)

            return True

        return False

    def _compare_directory(self, bp: StageBlueprint, existing_properties):
        is_alter_required = False

        # Check properties in blueprint against DESC output
        for prop_name, prop_value in coalesce(bp.directory, {}).items():
            # REFRESH_ON_CREATE is only used for CREATE command, but is not available for ALTER, must be skipped
            if prop_name == "REFRESH_ON_CREATE":
                continue

            if prop_name not in existing_properties["DIRECTORY"]:
                raise ValueError(f"Unknown DIRECTORY property [{prop_name}] for stage [{bp.full_name}]")

            if compare_dynamic_param_value(prop_value, existing_properties["DIRECTORY"][prop_name]["property_value"]):
                continue

            is_alter_required = True
            break

        # Check non-default properties in DESC output exist in blueprint
        for prop_name, prop in existing_properties["DIRECTORY"].items():
            # Informational properties, should not be compared or set explicitly
            if prop_name in ("LAST_REFRESHED_ON", "DIRECTORY_NOTIFICATION_CHANNEL"):
                continue

            if prop["property_value"] != prop["property_default"] and (prop_name not in coalesce(bp.directory, {})):
                is_alter_required = True
                break

        # If anything did not match, refresh all properties for this section
        if is_alter_required:
            query = self.engine.query_builder()
            query.append(
                "ALTER STAGE {full_name:i} SET",
                {
                    "full_name": bp.full_name,
                },
            )

            query.append_nl(self._build_directory(bp, existing_properties))
            self.engine.execute_safe_ddl(query)

        return is_alter_required

    def _compare_file_format(self, bp: StageBlueprint, existing_properties):
        # Only named FILE FORMATs are supported
        if "FORMAT_NAME" in existing_properties["STAGE_FILE_FORMAT"]:
            existing_file_format = str(existing_properties["STAGE_FILE_FORMAT"]["FORMAT_NAME"]["property_value"])
            existing_file_format = existing_file_format.replace("\\", "").replace('"', "")
        else:
            existing_file_format = None

        if not ((bp.file_format is None and existing_file_format is None) or str(bp.file_format) == existing_file_format):
            query = self.engine.query_builder()
            query.append(
                "ALTER STAGE {full_name:i} SET",
                {
                    "full_name": bp.full_name,
                },
            )

            query.append_nl(self._build_file_format(bp))
            self.engine.execute_safe_ddl(query)

            return True

        return False

    def _compare_copy_options(self, bp: StageBlueprint, existing_properties):
        is_alter_required = False

        # Check properties in blueprint against DESC output
        for prop_name, prop_value in coalesce(bp.copy_options, {}).items():
            if prop_name not in existing_properties["STAGE_COPY_OPTIONS"]:
                raise ValueError(f"Unknown COPY_OPTIONS property [{prop_name}] for stage [{bp.full_name}]")

            if compare_dynamic_param_value(prop_value, existing_properties["STAGE_COPY_OPTIONS"][prop_name]["property_value"]):
                continue

            is_alter_required = True
            break

        # Check non-default properties in DESC output exist in blueprint
        for prop_name, prop in existing_properties["STAGE_COPY_OPTIONS"].items():
            if prop["property_value"] != prop["property_default"] and (prop_name not in coalesce(bp.copy_options, {})):
                # ENFORCE_LENGTH is a mirror for TRUNCATECOLUMNS
                if prop_name == "ENFORCE_LENGTH" and "TRUNCATECOLUMNS" in coalesce(bp.copy_options, {}):
                    continue

                # TRUNCATECOLUMNS is a mirror for ENFORCE_LENGTH
                if prop_name == "TRUNCATECOLUMNS" and "ENFORCE_LENGTH" in coalesce(bp.copy_options, {}):
                    continue

                is_alter_required = True
                break

        # If anything did not match, refresh all properties for this section
        if is_alter_required:
            query = self.engine.query_builder()
            query.append(
                "ALTER STAGE {full_name:i} SET",
                {
                    "full_name": bp.full_name,
                },
            )

            query.append_nl(self._build_copy_options(bp))
            self.engine.execute_safe_ddl(query)

        return is_alter_required

    def _compare_comment(self, bp: StageBlueprint, row):
        if not ((bp.comment is None and row["comment"] is None) or str(bp.comment) == row["comment"]):
            query = self.engine.query_builder()
            query.append(
                "ALTER STAGE {full_name:i} SET",
                {
                    "full_name": bp.full_name,
                },
            )

            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

            self.engine.execute_safe_ddl(query)

            return True

        return False

    def _refresh_encryption(self, bp: StageBlueprint):
        if bp.encryption:
            query = self.engine.query_builder()
            query.append(
                "ALTER STAGE {full_name:i} SET",
                {
                    "full_name": bp.full_name,
                },
            )

            query.append_nl(self._build_encryption(bp))

            self.engine.execute_safe_ddl(query)

            return True

        return False

    def _build_directory(self, bp: StageBlueprint, existing_properties=None):
        query = self.engine.query_builder()

        if bp.directory:
            adjusted_directory = bp.directory

            # REFRESH_ON_CREATE is only used for CREATE command, but is not available for ALTER, must be skipped
            if existing_properties and "REFRESH_ON_CREATE" in adjusted_directory:
                del adjusted_directory["REFRESH_ON_CREATE"]

            if adjusted_directory:
                query.append("DIRECTORY = (")

                for k, v in adjusted_directory.items():
                    query.append(
                        "{param_name:r} = {param_value:dp}",
                        {
                            "param_name": k,
                            "param_value": v,
                        },
                    )

                query.append(")")
        else:
            query.append("DIRECTORY = ( ENABLE = FALSE )")

        return query

    def _build_file_format(self, bp: StageBlueprint):
        query = self.engine.query_builder()

        if bp.file_format:
            query.append(
                "FILE_FORMAT = ( FORMAT_NAME = {file_format:i} )",
                {
                    "file_format": bp.file_format,
                },
            )
        else:
            query.append("FILE_FORMAT = ( TYPE = CSV )")

        return query

    def _build_copy_options(self, bp: StageBlueprint):
        query = self.engine.query_builder()

        if bp.copy_options:
            query.append("COPY_OPTIONS = (")

            for k, v in bp.copy_options.items():
                query.append(
                    "{param_name:r} = {param_value:dp}",
                    {
                        "param_name": k,
                        "param_value": v,
                    },
                )

            query.append(")")
        else:
            query.append("COPY_OPTIONS = ()")

        return query

    def _build_encryption(self, bp: StageBlueprint):
        query = self.engine.query_builder()

        if bp.encryption:
            query.append("ENCRYPTION = (")

            for k, v in bp.encryption.items():
                query.append(
                    "{param_name:r} = {param_value:dp}",
                    {
                        "param_name": k,
                        "param_value": v,
                    },
                )

            query.append(")")

        # Cannot reset ENCRYPTION properties to default with ENCRYPTION = () or UNSET
        # It must be defined explicitly in config for any changes to take effect

        return query

    def _build_create_stage_sql(self, bp: StageBlueprint):
        query = self.engine.query_builder()

        if bp.url:
            query.append_nl("URL = {url}", {"url": bp.url})

        if bp.storage_integration:
            query.append_nl(
                "STORAGE_INTEGRATION = {storage_integration:i}",
                {
                    "storage_integration": bp.storage_integration,
                },
            )

        if bp.directory:
            query.append_nl(self._build_directory(bp))

        if bp.file_format:
            query.append_nl(self._build_file_format(bp))

        if bp.copy_options:
            query.append_nl(self._build_copy_options(bp))

        if bp.encryption:
            query.append_nl(self._build_encryption(bp))

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        return query

    def _get_existing_stage_properties(self, bp: StageBlueprint):
        cur = self.engine.execute_meta(
            "DESC STAGE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        result = {
            "STAGE_FILE_FORMAT": {},
            "STAGE_COPY_OPTIONS": {},
            "DIRECTORY": {},
        }

        # Add phantom parameters which should exist in output of DESC command
        # https://github.com/littleK0i/SnowDDL/issues/183
        result["DIRECTORY"]["AWS_SNS_TOPIC"] = {
            "parent_property": "DIRECTORY",
            "property": "AWS_SNS_TOPIC",
            "property_type": "String",
            "property_value": "",
            "property_default": "",
        }

        for r in cur:
            if r["parent_property"] not in result:
                result[r["parent_property"]] = {}

            result[r["parent_property"]][r["property"]] = r

        return result



from snowddl.blueprint import SemanticViewBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class SemanticViewResolver(AbstractSchemaObjectResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.SEMANTIC_VIEW

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW SEMANTIC VIEWS IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(SemanticViewBlueprint)

    def create_object(self, bp: SemanticViewBlueprint):
        create_query = self._build_create_semantic_view_sql(bp)

        # Temporary workaround due to not being able to change comment on existing SEMANTIC VIEW
        create_query.append_nl(
            "COMMENT = {comment}",
            {
                "comment": create_query.add_short_hash(bp.comment),
            },
        )

        self.engine.execute_safe_ddl(create_query)

        return ResolveResult.CREATE

    def compare_object(self, bp: SemanticViewBlueprint, row: dict):
        create_query = self._build_create_semantic_view_sql(bp)

        if not create_query.compare_short_hash(row["comment"]):
            # Temporary workaround due to not being able to change comment on existing SEMANTIC VIEW
            create_query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": create_query.add_short_hash(bp.comment),
                },
            )

            self.engine.execute_safe_ddl(create_query)

            return ResolveResult.REPLACE

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_safe_ddl(
            "DROP SEMANTIC VIEW {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _build_create_semantic_view_sql(self, bp: SemanticViewBlueprint):
        query = self.engine.query_builder()

        query.append(
            "CREATE OR REPLACE SEMANTIC VIEW {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        # ---

        query.append_nl("TABLES (")

        for idx, t in enumerate(bp.tables):
            query.append_nl(
                "  {comma:r}{table_alias:i} AS {table_name:i}",
                {
                    "comma": "  " if idx == 0 else ", ",
                    "table_alias": t.table_alias if t.table_alias else t.table_name.name,
                    "table_name": t.table_name,
                },
            )

            if t.primary_key:
                query.append(
                    "PRIMARY KEY ({col_names:i})",
                    {
                        "col_names": t.primary_key,
                    },
                )

            if t.unique:
                for unique_key in t.unique:
                    query.append(
                        "UNIQUE ({col_names:i})",
                        {
                            "col_names": unique_key,
                        },
                    )

            if t.with_synonyms:
                query.append(
                    "WITH SYNONYMS ({with_synonyms})",
                    {
                        "with_synonyms": t.with_synonyms,
                    },
                )

            if t.comment:
                query.append(
                    "COMMENT = {comment}",
                    {
                        "comment": t.comment,
                    },
                )

        query.append_nl(")")

        # ---

        if bp.relationships:
            query.append_nl("RELATIONSHIPS (")

            for idx, r in enumerate(bp.relationships):
                query.append_nl(
                    "  {comma:r}",
                    {
                        "comma": " " if idx == 0 else ",",
                    },
                )

                if r.relationship_identifier:
                    query.append(
                        "{relationship_identifier:i} AS",
                        {
                            "relationship_identifier": r.relationship_identifier,
                        },
                    )

                query.append(
                    "{table_alias:i} ({columns:i}) REFERENCES {ref_table_alias:i} ({ref_columns:i})",
                    {
                        "table_alias": r.table_alias,
                        "columns": r.columns,
                        "ref_table_alias": r.ref_table_alias,
                        "ref_columns": r.ref_columns,
                    },
                )

            query.append_nl(")")

        # ---

        if bp.facts:
            query.append_nl("FACTS (")

            for idx, e in enumerate(bp.facts):
                query.append_nl(
                    "  {comma:r}{table_alias:i}.{name:i} AS {sql:r}",
                    {
                        "comma": "  " if idx == 0 else ", ",
                        "table_alias": e.table_alias,
                        "name": e.name,
                        "sql": e.sql,
                    },
                )

                if e.with_synonyms:
                    query.append(
                        "WITH SYNONYMS ({with_synonyms})",
                        {
                            "with_synonyms": e.with_synonyms,
                        },
                    )

                if e.comment:
                    query.append(
                        "COMMENT = {comment}",
                        {
                            "comment": e.comment,
                        },
                    )

            query.append_nl(")")

        # ---

        if bp.dimensions:
            query.append_nl("DIMENSIONS (")

            for idx, e in enumerate(bp.dimensions):
                query.append_nl(
                    "  {comma:r}{table_alias:i}.{name:i} AS {sql:r}",
                    {
                        "comma": "  " if idx == 0 else ", ",
                        "table_alias": e.table_alias,
                        "name": e.name,
                        "sql": e.sql,
                    },
                )

                if e.with_synonyms:
                    query.append(
                        "WITH SYNONYMS ({with_synonyms})",
                        {
                            "with_synonyms": e.with_synonyms,
                        },
                    )

                if e.comment:
                    query.append(
                        "COMMENT = {comment}",
                        {
                            "comment": e.comment,
                        },
                    )

            query.append_nl(")")

        # ---

        if bp.metrics:
            query.append_nl("METRICS (")

            for idx, e in enumerate(bp.metrics):
                query.append_nl(
                    "  {comma:r}{table_alias:i}.{name:i} AS {sql:r}",
                    {
                        "comma": "  " if idx == 0 else ", ",
                        "table_alias": e.table_alias,
                        "name": e.name,
                        "sql": e.sql,
                    },
                )

                if e.with_synonyms:
                    query.append(
                        "WITH SYNONYMS ({with_synonyms})",
                        {
                            "with_synonyms": e.with_synonyms,
                        },
                    )

                if e.comment:
                    query.append(
                        "COMMENT = {comment}",
                        {
                            "comment": e.comment,
                        },
                    )

            query.append_nl(")")

        # ---

        # Comment is currently not fully supported due to lack of COMMENT ON SEMANTIC VIEW ... IS
        # As of Apr 2025, it is not possible to change comment of existing semantic view

        return query



from snowddl.blueprint import TechnicalRoleBlueprint
from snowddl.validator.abc_validator import AbstractValidator


class TechnicalRoleValidator(AbstractValidator):
    def get_blueprints(self):
        return self.config.get_blueprints_by_type(TechnicalRoleBlueprint)

    def validate_blueprint(self, bp: TechnicalRoleBlueprint):
        self._validate_grant_patterns(bp)
        self._validate_future_grant_patterns(bp)

    def _validate_grant_patterns(self, bp: TechnicalRoleBlueprint):
        for grant_pattern in bp.grant_patterns:
            if not self.config.get_blueprints_by_type_and_pattern(grant_pattern.on.blueprint_cls, grant_pattern.pattern):
                raise ValueError(
                    f"Technical role [{bp.full_name}] grant pattern [{grant_pattern.pattern}] "
                    f"does not match any objects of type [{grant_pattern.on.name}]"
                )

            if grant_pattern.privilege == "ALL":
                raise ValueError(
                    f"Technical role [{bp.full_name}] with grant privilege [{grant_pattern.on.name}:ALL] is not supported, "
                    f"each individual privilege must be defined explicitly, e.g. [TABLE:SELECT,UPDATE,DELETE]"
                )

            if grant_pattern.privilege == "OWNERSHIP":
                raise ValueError(
                    f"Technical role [{bp.full_name}] with grant privilege [{grant_pattern.on.name}:OWNERSHIP] is not supported, "
                    f"ownership belongs to database_owner, schema_owner or snowddl admin role"
                )

    def _validate_future_grant_patterns(self, bp: TechnicalRoleBlueprint):
        for future_grant_pattern in bp.future_grant_patterns:
            if not self.config.get_blueprints_by_type_and_pattern(
                future_grant_pattern.in_parent.blueprint_cls, future_grant_pattern.pattern
            ):
                raise ValueError(
                    f"Technical role [{bp.full_name}] future grant pattern [{future_grant_pattern.pattern}] "
                    f"does not match any objects of type [{future_grant_pattern.in_parent.name}]"
                )

            if future_grant_pattern.privilege == "ALL":
                raise ValueError(
                    f"Technical role [{bp.full_name}] with future grant privilege [{future_grant_pattern.on_future.name}:ALL] is not supported, "
                    f"each individual privilege must be defined explicitly, e.g. [TABLE:SELECT,UPDATE,DELETE]"
                )

            if future_grant_pattern.privilege == "OWNERSHIP":
                raise ValueError(
                    f"Technical role [{bp.full_name}] with future grant privilege [{future_grant_pattern.on_future.name}:OWNERSHIP] is not supported, "
                    f"ownership belongs to database_owner, schema_owner or snowddl admin role"
                )



from snowddl.blueprint import WarehouseBlueprint
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType, Edition
from snowddl.resolver._utils import compare_dynamic_param_value


class WarehouseResolver(AbstractResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.WAREHOUSE

    def get_existing_objects(self):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW WAREHOUSES LIKE {env_prefix:ls}",
            {
                "env_prefix": self.config.env_prefix,
            },
        )

        for r in cur:
            if r["owner"] != self.engine.context.current_role:
                continue

            existing_objects[r["name"]] = {
                "name": r["name"],
                "state": r["state"],
                "type": r["type"],
                "size": r["size"],
                "min_cluster_count": r.get("min_cluster_count", None),
                "max_cluster_count": r.get("max_cluster_count", None),
                "scaling_policy": r.get("scaling_policy", None),
                "auto_suspend": r["auto_suspend"],
                "resource_monitor": r["resource_monitor"] if r["resource_monitor"] != "null" else None,
                "enable_query_acceleration": r.get("enable_query_acceleration") == "true",
                "query_acceleration_max_scale_factor": r.get("query_acceleration_max_scale_factor"),
                "resource_constraint": r.get("resource_constraint"),
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(WarehouseBlueprint)

    def create_object(self, bp: WarehouseBlueprint):
        query = self.engine.query_builder()

        query.append(
            "CREATE WAREHOUSE {name:i}",
            {
                "name": bp.full_name,
            },
        )

        query.append_nl("WAREHOUSE_TYPE = {type}", {"type": bp.type})
        query.append_nl("WAREHOUSE_SIZE = {size}", {"size": self._normalise_warehouse_size(bp.size)})
        query.append_nl("AUTO_SUSPEND = {auto_suspend:d}", {"auto_suspend": bp.auto_suspend})
        query.append_nl("AUTO_RESUME = TRUE")
        query.append_nl("INITIALLY_SUSPENDED = TRUE")

        if bp.resource_constraint:
            query.append_nl("RESOURCE_CONSTRAINT = {resource_constraint}", {"resource_constraint": bp.resource_constraint})

        if self.engine.context.edition >= Edition.ENTERPRISE:
            query.append_nl("MIN_CLUSTER_COUNT = {min_cluster_count:d}", {"min_cluster_count": bp.min_cluster_count})
            query.append_nl("MAX_CLUSTER_COUNT = {max_cluster_count:d}", {"max_cluster_count": bp.max_cluster_count})
            query.append_nl("SCALING_POLICY = {scaling_policy}", {"scaling_policy": bp.scaling_policy})

            if bp.enable_query_acceleration:
                query.append_nl(
                    "ENABLE_QUERY_ACCELERATION = {enable_query_acceleration:b}",
                    {"enable_query_acceleration": bp.enable_query_acceleration},
                )

            if bp.query_acceleration_max_scale_factor:
                query.append_nl(
                    "QUERY_ACCELERATION_MAX_SCALE_FACTOR = {query_acceleration_max_scale_factor:d}",
                    {"query_acceleration_max_scale_factor": bp.query_acceleration_max_scale_factor},
                )

        query.append_nl("COMMENT = {comment}", {"comment": bp.comment})
        query.append_nl(self._build_common_parameters(bp))

        self.engine.execute_safe_ddl(query)

        if bp.resource_monitor:
            self.engine.execute_safe_ddl(
                "ALTER WAREHOUSE {full_name:i} SET RESOURCE_MONITOR = {resource_monitor:i}",
                {
                    "full_name": bp.full_name,
                    "resource_monitor": bp.resource_monitor,
                },
                condition=self.engine.settings.execute_resource_monitor,
            )

        return ResolveResult.CREATE

    def compare_object(self, bp: WarehouseBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        if self._compare_properties(bp, row):
            result = ResolveResult.ALTER

        if self._compare_resource_monitor(bp, row):
            result = ResolveResult.ALTER

        if self._compare_parameters(bp):
            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP WAREHOUSE {name:i}",
            {
                "name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _compare_properties(self, bp: WarehouseBlueprint, row: dict):
        query = self.engine.query_builder()

        query.append(
            "ALTER WAREHOUSE {name:i} SET",
            {
                "name": bp.full_name,
            },
        )

        if bp.type != row["type"]:
            query.append_nl("WAREHOUSE_TYPE = {type}", {"type": bp.type})

        if self._normalise_warehouse_size(bp.size) != self._normalise_warehouse_size(row["size"]):
            query.append_nl("WAREHOUSE_SIZE = {size}", {"size": self._normalise_warehouse_size(bp.size)})

        if bp.auto_suspend != row["auto_suspend"]:
            query.append_nl("AUTO_SUSPEND = {auto_suspend:d}", {"auto_suspend": bp.auto_suspend})

        if self._compare_resource_constraint(bp, row):
            query.append_nl("RESOURCE_CONSTRAINT = {resource_constraint}", {"resource_constraint": bp.resource_constraint})

        if self.engine.context.edition >= Edition.ENTERPRISE:
            if bp.min_cluster_count != row["min_cluster_count"]:
                query.append_nl("MIN_CLUSTER_COUNT = {min_cluster_count:d}", {"min_cluster_count": bp.min_cluster_count})

            if bp.max_cluster_count != row["max_cluster_count"]:
                query.append_nl("MAX_CLUSTER_COUNT = {max_cluster_count:d}", {"max_cluster_count": bp.max_cluster_count})

            if bp.scaling_policy != row["scaling_policy"]:
                query.append_nl("SCALING_POLICY = {scaling_policy}", {"scaling_policy": bp.scaling_policy})

            if bp.enable_query_acceleration != row["enable_query_acceleration"]:
                query.append_nl(
                    "ENABLE_QUERY_ACCELERATION = {enable_query_acceleration:b}",
                    {"enable_query_acceleration": bp.enable_query_acceleration},
                )

            if bp.query_acceleration_max_scale_factor != row["query_acceleration_max_scale_factor"]:
                query.append_nl(
                    "QUERY_ACCELERATION_MAX_SCALE_FACTOR = {query_acceleration_max_scale_factor:d}",
                    {"query_acceleration_max_scale_factor": bp.query_acceleration_max_scale_factor},
                )

        if bp.comment != row["comment"]:
            query.append_nl("COMMENT = {comment}", {"comment": bp.comment})

        if query.fragment_count() > 1:
            self.engine.execute_safe_ddl(query)
            return True

        return False

    def _compare_parameters(self, bp: WarehouseBlueprint):
        existing_params = self._get_existing_warehouse_parameters(bp)
        query = self.engine.query_builder()

        query.append(
            "ALTER WAREHOUSE {name:i} SET",
            {
                "name": bp.full_name,
            },
        )

        for param_name, param_value in bp.warehouse_params.items():
            if param_name not in existing_params:
                raise ValueError(f"Unknown parameter [{param_name}] for warehouse [{bp.full_name}]")

            if compare_dynamic_param_value(param_value, existing_params[param_name]["value"]):
                continue

            # At least one parameter in blueprint does not match the existing value
            # Refresh all parameters for such warehouse
            query.append(self._build_common_parameters(bp))
            break

        # If parameter was set on WAREHOUSE level and does not exist in blueprint anymore
        # Unset such parameter and reset it to default
        for param_name, p in existing_params.items():
            if p["level"] == "WAREHOUSE" and p["key"] not in bp.warehouse_params:
                # Setting parameter to NULL equals to UNSET
                query.append_nl(
                    "{param_name:r} = NULL",
                    {
                        "param_name": param_name,
                    },
                )

        if query.fragment_count() > 1:
            self.engine.execute_safe_ddl(query)
            return True

        return False

    def _compare_resource_monitor(self, bp: WarehouseBlueprint, row: dict):
        if bp.resource_monitor is None and row["resource_monitor"]:
            self.engine.execute_safe_ddl(
                "ALTER WAREHOUSE {full_name:i} UNSET RESOURCE_MONITOR",
                {
                    "full_name": bp.full_name,
                },
                condition=self.engine.settings.execute_resource_monitor,
            )

            return True

        if bp.resource_monitor and str(bp.resource_monitor) != row["resource_monitor"]:
            self.engine.execute_safe_ddl(
                "ALTER WAREHOUSE {full_name:i} SET RESOURCE_MONITOR = {resource_monitor:i}",
                {
                    "full_name": bp.full_name,
                    "resource_monitor": bp.resource_monitor,
                },
                condition=self.engine.settings.execute_resource_monitor,
            )

            return True

        return False

    def _build_common_parameters(self, bp: WarehouseBlueprint):
        query = self.engine.query_builder()

        for param_name, param_value in bp.warehouse_params.items():
            query.append_nl(
                "{param_name:r} = {param_value:dp}",
                {
                    "param_name": param_name,
                    "param_value": param_value,
                },
            )

        return query

    def _get_existing_warehouse_parameters(self, bp: WarehouseBlueprint):
        existing_params = {}

        cur = self.engine.execute_meta(
            "SHOW PARAMETERS IN WAREHOUSE {name:i}",
            {
                "name": bp.full_name,
            },
        )

        for r in cur:
            existing_params[r["key"]] = r

        return existing_params

    def _normalise_warehouse_size(self, size: str):
        return size.upper().replace("-", "")

    def _compare_resource_constraint(self, bp: WarehouseBlueprint, row: dict):
        if bp.resource_constraint == row["resource_constraint"]:
            return False

        # Special rules for RESOURCE_CONSTRAINT are required to take both old and new default logic into account
        if bp.resource_constraint is None and bp.type == "STANDARD" and row["resource_constraint"] == "STANDARD_GEN_1":
            return False

        return True

    def _post_process(self):
        if not self.engine.context.current_warehouse:
            return

        for result in self.resolved_objects.values():
            if result == ResolveResult.CREATE:
                # Revert current warehouse to original state if at least one object was created
                self.engine.execute_context_ddl(
                    "USE WAREHOUSE {full_name:i}", {"full_name": self.engine.context.current_warehouse}
                )

                break



from abc import abstractmethod
from typing import List, Optional, Tuple, Union

from snowddl.blueprint import (
    AccountGrant,
    AccountObjectIdent,
    ApplicationRoleIdent,
    DatabaseBlueprint,
    DatabaseRoleIdent,
    FutureGrant,
    Ident,
    IdentPattern,
    Grant,
    RoleBlueprint,
    SchemaBlueprint,
    SchemaIdent,
    build_role_ident,
    build_grant_name_ident,
    build_future_grant_name_ident,
)
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType


class AbstractRoleResolver(AbstractResolver):
    @abstractmethod
    def get_role_suffix(self) -> str:
        pass

    def get_role_type(self) -> Optional[str]:
        return None

    def get_role_pattern(self) -> Tuple[str, str]:
        role_type = self.get_role_type()
        role_suffix = self.get_role_suffix()

        if role_type:
            return (self.config.env_prefix, f"__{role_type}__{role_suffix}")

        return (self.config.env_prefix, f"__{role_suffix}")

    def get_object_type(self):
        return ObjectType.ROLE

    def get_existing_objects(self):
        existing_roles = {}

        cur = self.engine.execute_meta(
            "SHOW ROLES LIKE {pattern:lse}",
            {
                "pattern": self.get_role_pattern(),
            },
        )

        for r in cur:
            if r["owner"] != self.engine.context.current_role:
                continue

            existing_roles[r["name"]] = {
                "role_name": r["name"],
                "comment": r["comment"] if r["comment"] else None,
            }

        # Retrieve role grants in parallel
        for role_name, grants, account_grants, future_grants in self.engine.executor.map(
            self.get_existing_role_grants, existing_roles
        ):
            existing_roles[role_name]["grants"] = grants
            existing_roles[role_name]["account_grants"] = account_grants
            existing_roles[role_name]["future_grants"] = future_grants

        return existing_roles

    def get_existing_role_grants(self, role_name):
        grants = []
        account_grants = []
        future_grants = []

        # Normal and account grants
        cur = self.engine.execute_meta(
            "SHOW GRANTS TO ROLE {role_name:i}",
            {
                "role_name": role_name,
            },
        )

        for r in sorted(cur, key=self.sort_existing_grants):
            # Skip grants on unknown object types
            try:
                object_type = ObjectType[r["granted_on"]]
            except KeyError:
                continue

            # Snowflake bug: phantom MATERIALIZED VIEW when SEARCH OPTIMIZATION is enabled for table
            if object_type == ObjectType.MATERIALIZED_VIEW and str(r["name"]).endswith('IDX_MV_"'):
                continue

            if object_type == ObjectType.ACCOUNT:
                account_grants.append(AccountGrant(privilege=r["privilege"]))
            else:
                try:
                    grant_name = build_grant_name_ident(self.config.env_prefix, r["name"], object_type)
                except (KeyError, ValueError):
                    self.engine.intention_cache.add_object_name_warning(object_type, r["name"])
                    continue

                grants.append(
                    Grant(
                        privilege=r["privilege"],
                        on=object_type,
                        name=grant_name,
                    )
                )

        # Future grants
        cur = self.engine.execute_meta(
            "SHOW FUTURE GRANTS TO ROLE {role_name:i}",
            {
                "role_name": role_name,
            },
        )

        for r in sorted(cur, key=self.sort_existing_grants):
            try:
                object_type = ObjectType[r["grant_on"]]
            except KeyError:
                # Skip future grants on unknown object types
                continue

            try:
                grant_name = build_future_grant_name_ident(object_type, r["name"])
            except ValueError:
                self.engine.intention_cache.add_object_name_warning(object_type, r["name"])
                continue

            future_grants.append(
                FutureGrant(
                    privilege=r["privilege"],
                    on_future=object_type,
                    in_parent=ObjectType.SCHEMA if isinstance(grant_name, SchemaIdent) else ObjectType.DATABASE,
                    name=grant_name,
                )
            )

        return role_name, grants, account_grants, future_grants

    def sort_existing_grants(self, row):
        offset = 0

        # WRITE should be revoked before READ
        if row["privilege"] == "WRITE":
            offset = -1

        return (row.get("grant_on"), row.get("granted_on"), row["name"], offset, row["privilege"])

    def create_object(self, bp: RoleBlueprint):
        query = self.engine.query_builder()

        query.append(
            "CREATE ROLE {role_name:i}",
            {
                "role_name": bp.full_name,
            },
        )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        self.engine.execute_safe_ddl(query)

        self.engine.execute_safe_ddl(
            "GRANT ROLE {role_name:i} TO ROLE {current_role:i}",
            {
                "role_name": bp.full_name,
                "current_role": self.engine.context.current_role,
            },
        )

        for bp_grant in bp.grants:
            self.create_grant(bp.full_name, bp_grant)

        for bp_account_grant in bp.account_grants:
            self.create_account_grant(bp.full_name, bp_account_grant)

        for bp_future_grant in bp.future_grants:
            self.create_future_grant(bp.full_name, bp_future_grant)
            self.apply_future_grant_to_existing_objects(bp.full_name, bp_future_grant)

        return ResolveResult.CREATE

    def compare_object(self, bp: RoleBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        if bp.comment != row["comment"]:
            self.engine.execute_safe_ddl(
                "ALTER ROLE {role_name:i} SET COMMENT = {comment}",
                {
                    "role_name": bp.full_name,
                    "comment": bp.comment,
                },
            )

            result = ResolveResult.ALTER

        # Normal grants
        for existing_grant in row["grants"]:
            if existing_grant not in bp.grants and not any(fg.is_matching_grant(existing_grant) for fg in bp.future_grants):
                self.drop_grant(bp.full_name, existing_grant)
                result = ResolveResult.GRANT

        for bp_grant in bp.grants:
            if bp_grant not in row["grants"]:
                self.create_grant(bp.full_name, bp_grant)
                result = ResolveResult.GRANT

        # Account grants
        for existing_account_grant in row["account_grants"]:
            if existing_account_grant not in bp.account_grants:
                self.drop_account_grant(bp.full_name, existing_account_grant)
                result = ResolveResult.GRANT

        for bp_account_grant in bp.account_grants:
            if bp_account_grant not in row["account_grants"]:
                self.create_account_grant(bp.full_name, bp_account_grant)
                result = ResolveResult.GRANT

        # Future grants
        for existing_future_grant in row["future_grants"]:
            if existing_future_grant not in bp.future_grants:
                self.drop_future_grant(bp.full_name, existing_future_grant)
                result = ResolveResult.GRANT

        for bp_future_grant in bp.future_grants:
            if bp_future_grant not in row["future_grants"]:
                self.create_future_grant(bp.full_name, bp_future_grant)
                self.apply_future_grant_to_existing_objects(bp.full_name, bp_future_grant)
                result = ResolveResult.GRANT
            elif self.engine.settings.refresh_future_grants:
                self.apply_future_grant_to_existing_objects(bp.full_name, bp_future_grant)
                result = ResolveResult.GRANT

        return result

    def drop_object(self, row: dict):
        # All other grants and future grants are destroyed automatically
        self.engine.execute_unsafe_ddl(
            "DROP ROLE {role_name:i}",
            {
                "role_name": row["role_name"],
            },
        )

        return ResolveResult.DROP

    def create_grant(self, role_name, grant: Grant):
        if grant.privilege == "USAGE" and grant.on in (ObjectType.ROLE, ObjectType.APPLICATION_ROLE, ObjectType.DATABASE_ROLE):
            self.engine.execute_safe_ddl(
                "GRANT {on:r} {name:i} TO ROLE {role_name:i}",
                {
                    "on": grant.on.singular_for_grant,
                    "name": grant.name,
                    "role_name": role_name,
                },
            )
        else:
            self.engine.execute_safe_ddl(
                "GRANT {privilege:r} ON {on:r} {name:i} TO ROLE {role_name:i}",
                {
                    "privilege": grant.privilege,
                    "on": grant.on.singular_for_grant,
                    "name": grant.name,
                    "role_name": role_name,
                },
            )

    def drop_grant(self, role_name, grant: Grant):
        # OWNERSHIP can only be transferred, not revoked
        # We transfer ownership back to default "SnowDDL admin" role instead of REVOKE
        if grant.privilege == "OWNERSHIP":
            # Changing ownership of notebooks and shares is not supported by Snowflake
            # https://docs.snowflake.com/en/user-guide/ui-snowsight/notebooks-limitations
            # https://docs.snowflake.com/en/sql-reference/sql/grant-ownership#usage-notes
            if grant.on in (ObjectType.NOTEBOOK, ObjectType.SHARE):
                return

            self.engine.execute_safe_ddl(
                "GRANT {privilege:r} ON {on:r} {name:i} TO ROLE {current_role:i} COPY CURRENT GRANTS",
                {
                    "privilege": grant.privilege,
                    "on": grant.on.singular_for_grant,
                    "name": grant.name,
                    "current_role": self.engine.context.current_role,
                },
            )
        elif grant.privilege == "USAGE" and grant.on in (ObjectType.ROLE, ObjectType.APPLICATION_ROLE, ObjectType.DATABASE_ROLE):
            self.engine.execute_safe_ddl(
                "REVOKE {on:r} {name:i} FROM ROLE {role_name:i}",
                {
                    "on": grant.on.singular_for_grant,
                    "name": grant.name,
                    "role_name": role_name,
                },
            )
        else:
            self.engine.execute_safe_ddl(
                "REVOKE {privilege:r} ON {on:r} {name:i} FROM ROLE {role_name:i}",
                {
                    "privilege": grant.privilege,
                    "on": grant.on.singular_for_grant,
                    "name": grant.name,
                    "role_name": role_name,
                },
            )

    def create_account_grant(self, role_name, account_grant: AccountGrant):
        self.engine.execute_safe_ddl(
            "GRANT {privilege:r} ON ACCOUNT TO ROLE {role_name:i}",
            {
                "privilege": account_grant.privilege,
                "role_name": role_name,
            },
        )

    def drop_account_grant(self, role_name, account_grant: AccountGrant):
        self.engine.execute_safe_ddl(
            "REVOKE {privilege:r} ON ACCOUNT FROM ROLE {role_name:i}",
            {
                "privilege": account_grant.privilege,
                "role_name": role_name,
            },
        )

    def create_future_grant(self, role_name, grant: FutureGrant):
        self.engine.execute_safe_ddl(
            "GRANT {privilege:r} ON FUTURE {on_future_plural:r} IN {in_parent_singular:r} {name:i} TO ROLE {role_name:i}",
            {
                "privilege": grant.privilege,
                "on_future_plural": grant.on_future.plural,
                "in_parent_singular": grant.in_parent.singular,
                "name": grant.name,
                "role_name": role_name,
            },
        )

    def drop_future_grant(self, role_name, grant: FutureGrant):
        self.engine.execute_safe_ddl(
            "REVOKE {privilege:r} ON FUTURE {on_future_plural:r} IN {in_parent_singular:r} {name:i} FROM ROLE {role_name:i}",
            {
                "privilege": grant.privilege,
                "on_future_plural": grant.on_future.plural,
                "in_parent_singular": grant.in_parent.singular,
                "name": grant.name,
                "role_name": role_name,
            },
        )

    def apply_future_grant_to_existing_objects(self, role_name, grant: FutureGrant):
        # Bulk grant on objects of type PIPE to ROLE is restricted (by Snowflake)
        if grant.on_future == ObjectType.PIPE:
            return

        self.engine.execute_safe_ddl(
            "GRANT {privilege:r} ON ALL {on_future_plural:r} IN {in_parent_singular:r} {name:i} TO ROLE {role_name:i}{copy_grants:r}",
            {
                "privilege": grant.privilege,
                "on_future_plural": grant.on_future.plural,
                "in_parent_singular": grant.in_parent.singular,
                "name": grant.name,
                "role_name": role_name,
                "copy_grants": " COPY CURRENT GRANTS" if (grant.privilege == "OWNERSHIP") else "",
            },
        )

    def build_database_role_grants(self, database_name_pattern: IdentPattern, role_type: str) -> List[Grant]:
        grants = []

        for database_bp in self.config.get_blueprints_by_type_and_pattern(DatabaseBlueprint, database_name_pattern).values():
            database_permission_model = self.config.get_permission_model(database_bp.permission_model)

            if database_permission_model.ruleset.create_database_owner_role:
                # Databases with "database owner" permission model are assigned directly
                grants.append(
                    Grant(
                        privilege="USAGE",
                        on=ObjectType.ROLE,
                        name=build_role_ident(
                            self.config.env_prefix,
                            database_bp.full_name.database,
                            role_type,
                            self.config.DATABASE_ACCESS_ROLE_SUFFIX,
                        ),
                    )
                )

            elif database_permission_model.ruleset.create_schema_owner_role:
                # Databases with "schema owner" permission model are automatically expanded into individual schema roles
                schema_name_pattern = IdentPattern(f"{database_bp.full_name.database}.*")
                grants.extend(self.build_schema_role_grants(schema_name_pattern, role_type))

        return grants

    def build_schema_role_grants(self, schema_name_pattern: IdentPattern, role_type: str) -> List[Grant]:
        grants = []

        for schema_bp in self.config.get_blueprints_by_type_and_pattern(SchemaBlueprint, schema_name_pattern).values():
            grants.append(
                Grant(
                    privilege="USAGE",
                    on=ObjectType.ROLE,
                    name=build_role_ident(
                        self.config.env_prefix,
                        schema_bp.full_name.database,
                        schema_bp.full_name.schema,
                        role_type,
                        self.config.SCHEMA_ACCESS_ROLE_SUFFIX,
                    ),
                )
            )

        return grants

    def build_warehouse_role_grant(self, warehouse_name: AccountObjectIdent, role_type) -> Grant:
        return Grant(
            privilege="USAGE",
            on=ObjectType.ROLE,
            name=build_role_ident(
                self.config.env_prefix, warehouse_name.name, role_type, self.config.WAREHOUSE_ACCESS_ROLE_SUFFIX
            ),
        )

    def build_share_read_grant(self, share_name: Union[Ident, DatabaseRoleIdent]) -> Grant:
        if isinstance(share_name, DatabaseRoleIdent):
            return Grant(
                privilege="USAGE",
                on=ObjectType.DATABASE_ROLE,
                name=share_name,
            )
        else:
            return Grant(
                privilege="USAGE",
                on=ObjectType.ROLE,
                name=build_role_ident(self.config.env_prefix, share_name, self.config.SHARE_ACCESS_ROLE_SUFFIX),
            )

    def build_application_role_grant(self, application_role_name: ApplicationRoleIdent) -> Grant:
        return Grant(
            privilege="USAGE",
            on=ObjectType.APPLICATION_ROLE,
            name=application_role_name,
        )

    def build_technical_role_grant(self, technical_role_name: AccountObjectIdent) -> Grant:
        return Grant(
            privilege="USAGE",
            on=ObjectType.ROLE,
            name=technical_role_name,
        )

    def build_global_role_grant(self, global_role_name: Ident) -> Grant:
        return Grant(
            privilege="USAGE",
            on=ObjectType.ROLE,
            name=global_role_name,
        )

    def build_integration_grant(self, integration_name: Ident) -> Grant:
        return Grant(
            privilege="USAGE",
            on=ObjectType.INTEGRATION,
            name=integration_name,
        )



from snowddl.blueprint import MaterializedViewBlueprint, Edition
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class MaterializedViewResolver(AbstractSchemaObjectResolver):
    skip_min_edition = Edition.ENTERPRISE

    def get_object_type(self) -> ObjectType:
        return ObjectType.MATERIALIZED_VIEW

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW MATERIALIZED VIEWS IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                "text": str(r["text"]).rstrip(";"),
                "is_secure": r["is_secure"] == "true",
                "cluster_by": r["cluster_by"] if r["cluster_by"] else None,
                "invalid": r["invalid"] == "true",
                "invalid_reason": r["invalid_reason"],
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(MaterializedViewBlueprint)

    def create_object(self, bp: MaterializedViewBlueprint):
        self.engine.execute_safe_ddl(self._build_create_materialized_view(bp))

        return ResolveResult.CREATE

    def compare_object(self, bp: MaterializedViewBlueprint, row: dict):
        query = self._build_create_materialized_view(bp)

        # If view text is exactly the same, and view is still valid, do nothing
        if row["text"] == str(query) and not row["invalid"]:
            return ResolveResult.NOCHANGE

        self.engine.execute_unsafe_ddl(query)

        return ResolveResult.REPLACE

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP MATERIALIZED VIEW {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _build_create_materialized_view(self, bp: MaterializedViewBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE OR REPLACE")

        if bp.is_secure:
            query.append("SECURE")

        query.append(
            "MATERIALIZED VIEW {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        query.append_nl("COPY GRANTS")

        if bp.columns:
            query.append_nl("(")

            for idx, c in enumerate(bp.columns):
                query.append_nl(
                    "    {comma:r}{col_name:i}",
                    {
                        "comma": "  " if idx == 0 else ", ",
                        "col_name": c.name,
                    },
                )

                if c.comment:
                    query.append(
                        "COMMENT {col_comment}",
                        {
                            "col_comment": c.comment,
                        },
                    )

            query.append_nl(")")

        if bp.cluster_by:
            query.append_nl(
                "CLUSTER BY ({cluster_by:r})",
                {
                    "cluster_by": bp.cluster_by,
                },
            )

        query.append_nl("AS")
        query.append_nl(bp.text)

        return query



from snowddl.blueprint import RoleBlueprint, UserBlueprint, Grant, build_role_ident, build_grant_name_ident
from snowddl.resolver.abc_role_resolver import AbstractRoleResolver, ObjectType


class UserRoleResolver(AbstractRoleResolver):
    def get_role_suffix(self):
        return self.config.USER_ROLE_SUFFIX

    def get_existing_role_grants(self, role_name):
        grants = []

        cur = self.engine.execute_meta(
            "SHOW GRANTS TO ROLE {role_name:i}",
            {
                "role_name": role_name,
            },
        )

        for r in cur:
            # Check ROLE grants only, ignore everything else
            # User roles may accumulate random grants from temporary tables and stages
            # as well as ownership of manually created objects
            if r["granted_on"] != "ROLE":
                continue

            object_type = ObjectType.ROLE

            try:
                grant_name = build_grant_name_ident(self.config.env_prefix, r["name"], object_type)
            except (KeyError, ValueError):
                self.engine.intention_cache.add_object_name_warning(object_type, r["name"])
                continue

            grants.append(
                Grant(
                    privilege=r["privilege"],
                    on=ObjectType[r["granted_on"]],
                    name=grant_name,
                )
            )

        return role_name, grants, [], []

    def get_blueprints(self):
        blueprints = []

        for user in self.config.get_blueprints_by_type(UserBlueprint).values():
            blueprints.append(self.get_blueprint_user_role(user))

        return {str(bp.full_name): bp for bp in blueprints}

    def get_blueprint_user_role(self, user: UserBlueprint):
        grants = []

        for business_role in user.business_roles:
            grants.append(
                Grant(
                    privilege="USAGE",
                    on=ObjectType.ROLE,
                    name=business_role,
                )
            )

        bp = RoleBlueprint(
            full_name=build_role_ident(self.config.env_prefix, user.full_name.name, self.get_role_suffix()),
            grants=grants,
        )

        return bp



from ._yaml import SnowDDLDumper, YamlFoldedStr, YamlLiteralStr
from .abc_converter import AbstractConverter, ConvertResult
from .abc_schema_object_converter import AbstractSchemaObjectConverter
from .database import DatabaseConverter
from .schema import SchemaConverter
from .sequence import SequenceConverter
from .table import TableConverter
from .task import TaskConverter
from .view import ViewConverter
from .function import FunctionConverter

default_converter_sequence = [
    DatabaseConverter,
    SchemaConverter,
    SequenceConverter,
    TableConverter,
    TaskConverter,
    ViewConverter,
    FunctionConverter,
]



from yaml import SafeDumper, add_representer


class SnowDDLDumper(SafeDumper):
    pass


class YamlFoldedStr(str):
    pass


class YamlLiteralStr(str):
    pass


class YamlIncludeStr(str):
    pass


def folded_str_presenter(dumper: SnowDDLDumper, data):
    return dumper.represent_scalar("tag:yaml.org,2002:str", data, style=">")


def literal_str_presenter(dumper: SnowDDLDumper, data):
    return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")


def include_str_presenter(dumper: SnowDDLDumper, data):
    return dumper.represent_scalar("!include", data)


add_representer(YamlFoldedStr, folded_str_presenter, Dumper=SnowDDLDumper)
add_representer(YamlLiteralStr, literal_str_presenter, Dumper=SnowDDLDumper)
add_representer(YamlIncludeStr, include_str_presenter, Dumper=SnowDDLDumper)



from snowddl.blueprint import (
    MaskingPolicyBlueprint,
    Ident,
    SchemaObjectIdent,
    NameWithType,
    DataType,
    MaskingPolicyReference,
    ObjectType,
    build_schema_object_ident,
)
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
masking_policy_json_schema = {
    "type": "object",
    "properties": {
        "arguments": {
            "type": "object",
            "additionalProperties": {
                "type": "string"
            }
        },
        "returns": {
            "type": "string"
        },
        "body": {
            "type": "string"
        },
        "exempt_other_policies": {
            "type": "boolean"
        },
        "references": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "object_type": {
                        "type": "string"
                    },
                    "object_name": {
                        "type": "string"
                    },
                    "columns": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    }
                },
                "required": ["object_type", "object_name", "columns"],
                "additionalProperties": False
            },
            "minItems": 1
        },
        "comment": {
            "type": "string"
        }
    },
    "required": ["arguments", "returns", "body"],
    "additionalProperties": False
}
# fmt: on


class MaskingPolicyParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("masking_policy", masking_policy_json_schema, self.process_masking_policy)

    def process_masking_policy(self, f: ParsedFile):
        arguments = [NameWithType(name=Ident(k), type=DataType(t)) for k, t in f.params.get("arguments", {}).items()]
        references = []

        for a in f.params.get("references", []):
            ref = MaskingPolicyReference(
                object_type=ObjectType[a["object_type"].upper()],
                object_name=build_schema_object_ident(self.env_prefix, a["object_name"], f.database, f.schema),
                columns=[Ident(c) for c in a["columns"]],
            )

            references.append(ref)

        bp = MaskingPolicyBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            body=self.normalise_sql_text_param(f.params["body"]),
            arguments=arguments,
            returns=DataType(f.params["returns"]),
            exempt_other_policies=f.params.get("exempt_other_policies", False),
            references=references,
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)



from abc import ABC, abstractmethod
from enum import Enum
from concurrent.futures import as_completed
from jsonschema import validate
from pathlib import Path
from traceback import format_exc
from typing import Dict, List, TYPE_CHECKING
from yaml import dump_all

from snowddl.error import SnowDDLExecuteError
from snowddl.blueprint import ObjectType, Edition
from snowddl.converter._yaml import SnowDDLDumper

if TYPE_CHECKING:
    from snowddl.engine import SnowDDLEngine


class ConvertResult(Enum):
    DUMP = "DUMP"
    EMPTY = "EMPTY"
    SKIP = "SKIP"
    ERROR = "ERROR"


class AbstractConverter(ABC):
    skip_min_edition = Edition.STANDARD

    def __init__(self, engine: "SnowDDLEngine", base_path: Path):
        self.engine = engine
        self.config = engine.config

        self.base_path = base_path
        self.env_prefix = self.config.env_prefix

        self.object_type = self.get_object_type()
        self.existing_objects: Dict[str, Dict] = {}

        self.converted_objects: Dict[ConvertResult, List[str]] = {k: [] for k in ConvertResult}
        self.errors: Dict[str, Exception] = {}

    def convert(self):
        if self._is_skipped():
            return

        self._check_optional_dependencies()

        try:
            self.existing_objects = self.get_existing_objects()
        except SnowDDLExecuteError as e:
            self.engine.logger.info(
                f"Could not get existing objects for converter [{self.__class__.__name__}]: \n{e.verbose_message()}"
            )
            raise e.snow_exc

        tasks = {}

        for full_name in sorted(self.existing_objects):
            tasks[full_name] = (self.dump_object, self.existing_objects[full_name])

        self._process_tasks(tasks)

    def _process_tasks(self, tasks):
        futures = {}

        for full_name, args in tasks.items():
            futures[self.engine.executor.submit(*args)] = full_name

        for f in as_completed(futures):
            full_name = futures[f]

            try:
                result = f.result()
                self.engine.logger.info(f"Converted {self.object_type.name} [{full_name}]: {result.value}")
            except Exception as e:
                result = ConvertResult.ERROR

                if isinstance(e, SnowDDLExecuteError):
                    error_text = e.verbose_message()
                else:
                    error_text = format_exc()

                self.engine.logger.warning(f"Converted {self.object_type.name} [{full_name}]: {result.value}\n{error_text}")
                self.errors[full_name] = e

            self.converted_objects[result].append(full_name)

    def _is_skipped(self):
        if self.engine.context.edition < self.skip_min_edition:
            return True

        if self.object_type in self.engine.settings.exclude_object_types:
            return True

        if self.engine.settings.include_object_types:
            return self.object_type not in self.engine.settings.include_object_types

        return False

    def _check_optional_dependencies(self):
        pass

    def _dump_file(self, file_path: Path, data: Dict, json_schema: dict):
        # Remove None values
        data = {k: v for k, v in data.items() if v is not None}

        # Validate JSON schema
        validate(data, json_schema)

        # Create directory if not exist
        file_path.parent.mkdir(mode=0o755, parents=True, exist_ok=True)

        # (Over)write file
        with file_path.open("w", encoding="utf-8") as f:
            dump_all([data], f, Dumper=SnowDDLDumper, sort_keys=False)

    def _dump_code(self, file_path: Path, content: str):
        # Create directory if not exist
        file_path.parent.mkdir(mode=0o755, parents=True, exist_ok=True)

        # Write data as plain text
        file_path.write_text(content)

    def _normalise_name(self, name: str):
        return name.lower()

    def _normalise_name_with_prefix(self, name: str):
        return self._normalise_name(name.removeprefix(self.env_prefix))

    @abstractmethod
    def get_object_type(self) -> ObjectType:
        pass

    @abstractmethod
    def get_existing_objects(self) -> Dict[str, Dict]:
        pass

    @abstractmethod
    def dump_object(self, row: Dict) -> ConvertResult:
        pass



from snowddl.blueprint import DatabaseBlueprint, SchemaBlueprint
from snowddl.validator.abc_validator import AbstractValidator


class DatabaseValidator(AbstractValidator):
    def get_blueprints(self):
        return self.config.get_blueprints_by_type(DatabaseBlueprint)

    def validate_blueprint(self, bp: DatabaseBlueprint):
        self._validate_owner_parameters(bp)
        self._validate_owner_patterns(bp)

    def _validate_owner_parameters(self, bp: DatabaseBlueprint):
        database_permission_model = self.config.get_permission_model(bp.permission_model)

        if not database_permission_model.ruleset.create_database_owner_role:
            if bp.owner_database_write:
                raise ValueError(
                    f"Cannot use parameter owner_database_write for database [{bp.full_name}] due to permission model ruleset"
                )

            if bp.owner_database_read:
                raise ValueError(
                    f"Cannot use parameter owner_database_read for database [{bp.full_name}] due to permission model ruleset"
                )

            if bp.owner_schema_write:
                raise ValueError(
                    f"Cannot use parameter owner_schema_write for database [{bp.full_name}] due to permission model ruleset"
                )

            if bp.owner_schema_read:
                raise ValueError(
                    f"Cannot use parameter owner_schema_read for database [{bp.full_name}] due to permission model ruleset"
                )

            if bp.owner_integration_usage:
                raise ValueError(
                    f"Cannot use parameter owner_integration_usage for database [{bp.full_name}] due to permission model ruleset"
                )

            if bp.owner_share_read:
                raise ValueError(
                    f"Cannot use parameter owner_share_read for database [{bp.full_name}] due to permission model ruleset"
                )

            if bp.owner_warehouse_usage:
                raise ValueError(
                    f"Cannot use parameter owner_warehouse_usage for database [{bp.full_name}] due to permission model ruleset"
                )

            if bp.owner_account_grants:
                raise ValueError(
                    f"Cannot use parameter owner_account_grants for database [{bp.full_name}] due to permission model ruleset"
                )

            if bp.owner_global_roles:
                raise ValueError(
                    f"Cannot use parameter owner_global_roles for database [{bp.full_name}] due to permission model ruleset"
                )

    def _validate_owner_patterns(self, bp: DatabaseBlueprint):
        for database_name_pattern in bp.owner_database_write:
            if not self.config.get_blueprints_by_type_and_pattern(DatabaseBlueprint, database_name_pattern):
                raise ValueError(
                    f"Database [{bp.full_name}] owner_database_write pattern [{database_name_pattern}] "
                    f"does not match any databases"
                )

        for database_name_pattern in bp.owner_database_read:
            if not self.config.get_blueprints_by_type_and_pattern(DatabaseBlueprint, database_name_pattern):
                raise ValueError(
                    f"Database [{bp.full_name}] owner_database_read pattern [{database_name_pattern}] "
                    f"does not match any databases"
                )

        for schema_name_pattern in bp.owner_schema_write:
            if not self.config.get_blueprints_by_type_and_pattern(SchemaBlueprint, schema_name_pattern):
                raise ValueError(
                    f"Database [{bp.full_name}] owner_schema_write pattern [{schema_name_pattern}] " f"does not match any schemas"
                )

        for schema_name_pattern in bp.owner_schema_read:
            if not self.config.get_blueprints_by_type_and_pattern(SchemaBlueprint, schema_name_pattern):
                raise ValueError(
                    f"Database [{bp.full_name}] owner_schema_read pattern [{schema_name_pattern}] " f"does not match any schemas"
                )



from snowddl.blueprint import ResourceMonitorBlueprint, AccountObjectIdent
from snowddl.parser.abc_parser import AbstractParser


# fmt: off
resource_monitor_json_schema = {
    "type": "object",
    "additionalProperties": {
        "type": "object",
        "properties": {
            "credit_quota": {
                "type": "integer"
            },
            "frequency": {
                "type": "string"
            },
            "triggers": {
                "type": "object",
                "additionalProperties": {
                    "type": "string"
                }
            }
        },
        "required": ["credit_quota", "frequency", "triggers"],
        "additionalProperties": False
    }
}
# fmt: on


class ResourceMonitorParser(AbstractParser):
    def load_blueprints(self):
        self.parse_multi_entity_file("resource_monitor", resource_monitor_json_schema, self.process_resource_monitor)

    def process_resource_monitor(self, resource_monitor_name, resource_monitor_params):
        bp = ResourceMonitorBlueprint(
            full_name=AccountObjectIdent(self.env_prefix, resource_monitor_name),
            credit_quota=int(resource_monitor_params.get("credit_quota")),
            frequency=str(resource_monitor_params.get("frequency")).upper(),
            triggers={int(k): str(v).upper() for k, v in resource_monitor_params.get("triggers").items()},
            comment=None,
        )

        self.config.add_blueprint(bp)



from json import loads

from snowddl.blueprint import RowAccessPolicyBlueprint, ObjectType, Edition, SchemaObjectIdent
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult


class RowAccessPolicyResolver(AbstractSchemaObjectResolver):
    skip_on_empty_blueprints = True
    skip_min_edition = Edition.ENTERPRISE

    def get_object_type(self) -> ObjectType:
        return ObjectType.ROW_ACCESS_POLICY

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW ROW ACCESS POLICIES IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            full_name = f"{r['database_name']}.{r['schema_name']}.{r['name']}"

            existing_objects[full_name] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(RowAccessPolicyBlueprint)

    def create_object(self, bp: RowAccessPolicyBlueprint):
        self._create_policy(bp)
        self._apply_policy_refs(bp, skip_existing=True)

        return ResolveResult.CREATE

    def compare_object(self, bp: RowAccessPolicyBlueprint, row: dict):
        cur = self.engine.execute_meta(
            "DESC ROW ACCESS POLICY {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        r = cur.fetchone()

        # If signature was changed, policy and all references must be dropped and created again
        if r["signature"] != f"({', '.join([f'{a.name} {a.type.base_type.name}' for a in bp.arguments])})":
            self._drop_policy_refs(bp.full_name)
            self._drop_policy(bp.full_name)

            self._create_policy(bp)
            self._apply_policy_refs(bp, skip_existing=True)

            return ResolveResult.REPLACE

        result = ResolveResult.NOCHANGE

        if self._apply_policy_refs(bp):
            result = ResolveResult.ALTER

        if r["body"] != bp.body:
            self.engine.execute_unsafe_ddl(
                "ALTER ROW ACCESS POLICY {full_name:i} SET BODY -> {body:r}",
                {
                    "full_name": bp.full_name,
                    "body": bp.body,
                },
                condition=self.engine.settings.execute_row_access_policy,
            )

            result = ResolveResult.ALTER

        if row["comment"] != bp.comment:
            self.engine.execute_unsafe_ddl(
                "ALTER ROW ACCESS POLICY {full_name:i} SET COMMENT = {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": bp.comment,
                },
                condition=self.engine.settings.execute_row_access_policy,
            )

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self._drop_policy_refs(SchemaObjectIdent("", row["database"], row["schema"], row["name"]))
        self._drop_policy(SchemaObjectIdent("", row["database"], row["schema"], row["name"]))

        return ResolveResult.DROP

    def _create_policy(self, bp: RowAccessPolicyBlueprint):
        query = self.engine.query_builder()

        query.append(
            "CREATE ROW ACCESS POLICY {full_name:i} AS (",
            {
                "full_name": bp.full_name,
            },
        )

        for idx, arg in enumerate(bp.arguments):
            query.append_nl(
                "    {comma:r}{arg_name:i} {arg_type:r}",
                {
                    "comma": "  " if idx == 0 else ", ",
                    "arg_name": arg.name,
                    "arg_type": arg.type,
                },
            )

        query.append_nl(")")

        query.append_nl("RETURNS BOOLEAN -> ")

        query.append_nl(
            "{body:r}",
            {
                "body": bp.body,
            },
        )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_row_access_policy)

    def _drop_policy(self, policy_name: SchemaObjectIdent):
        self.engine.execute_unsafe_ddl(
            "DROP ROW ACCESS POLICY {full_name:i}",
            {
                "full_name": policy_name,
            },
            condition=self.engine.settings.execute_row_access_policy,
        )

    def _apply_policy_refs(self, bp: RowAccessPolicyBlueprint, skip_existing=False):
        existing_policy_refs = {} if skip_existing else self._get_existing_policy_refs(bp.full_name)
        applied_change = False

        for ref in bp.references:
            ref_key = f"{ref.object_type.name}|{ref.object_name}|{bp.full_name}"

            # Policy was applied before
            if ref_key in existing_policy_refs:
                existing_ref_columns = existing_policy_refs[ref_key]["columns"]
                del existing_policy_refs[ref_key]

                # Policy was applied to exactly the same columns
                if [str(c) for c in ref.columns] == existing_ref_columns:
                    continue

                # Policy was applied to different columns, should be dropped and re-created
                self.engine.execute_unsafe_ddl(
                    "ALTER {object_type:r} {object_name:i} DROP ROW ACCESS POLICY {policy_name:i}",
                    {
                        "object_type": ref.object_type.singular_for_ref,
                        "object_name": ref.object_name,
                        "policy_name": bp.full_name,
                    },
                    condition=self.engine.settings.execute_row_access_policy,
                )

            # Apply new policy
            self.engine.execute_unsafe_ddl(
                "ALTER {object_type:r} {object_name:i} ADD ROW ACCESS POLICY {policy_name:i} ON ({columns:i})",
                {
                    "object_type": ref.object_type.singular_for_ref,
                    "object_name": ref.object_name,
                    "policy_name": bp.full_name,
                    "columns": ref.columns,
                },
                condition=self.engine.settings.execute_row_access_policy,
            )

            applied_change = True

        # Remove remaining policy references which no longer exist in blueprint
        for existing_ref in existing_policy_refs.values():
            self.engine.execute_unsafe_ddl(
                "ALTER {object_type:r} {database:i}.{schema:i}.{name:i} DROP ROW ACCESS POLICY {policy_name:i}",
                {
                    "object_type": ObjectType[existing_ref["object_type"]].singular_for_ref,
                    "database": existing_ref["database"],
                    "schema": existing_ref["schema"],
                    "name": existing_ref["name"],
                    "policy_name": bp.full_name,
                },
                condition=self.engine.settings.execute_row_access_policy,
            )

            applied_change = True

        return applied_change

    def _drop_policy_refs(self, policy_name: SchemaObjectIdent):
        existing_policy_refs = self._get_existing_policy_refs(policy_name)

        for existing_ref in existing_policy_refs.values():
            self.engine.execute_unsafe_ddl(
                "ALTER {object_type:r} {database:i}.{schema:i}.{name:i} DROP ROW ACCESS POLICY {policy_name:i}",
                {
                    "object_type": ObjectType[existing_ref["object_type"]].singular_for_ref,
                    "database": existing_ref["database"],
                    "schema": existing_ref["schema"],
                    "name": existing_ref["name"],
                    "policy_name": policy_name,
                },
                condition=self.engine.settings.execute_row_access_policy,
            )

    def _get_existing_policy_refs(self, policy_name: SchemaObjectIdent):
        existing_policy_refs = {}

        cur = self.engine.execute_meta(
            "SELECT * FROM TABLE(snowflake.information_schema.policy_references(policy_name => {policy_name}))",
            {
                "policy_name": policy_name,
            },
        )

        for r in cur:
            ref_key = (
                f"{r['REF_ENTITY_DOMAIN']}"
                f"|{r['REF_DATABASE_NAME']}.{r['REF_SCHEMA_NAME']}.{r['REF_ENTITY_NAME']}"
                f"|{r['POLICY_DB']}.{r['POLICY_SCHEMA']}.{r['POLICY_NAME']}"
            )

            existing_policy_refs[ref_key] = {
                "object_type": r["REF_ENTITY_DOMAIN"],
                "database": r["REF_DATABASE_NAME"],
                "schema": r["REF_SCHEMA_NAME"],
                "name": r["REF_ENTITY_NAME"],
                "columns": loads(r["REF_ARG_COLUMN_NAMES"]),
            }

        return existing_policy_refs
